__all__ = ["Client"]

import datetime
import json
from typing import Any, Literal

from unicex._base import BaseClient
from unicex.exceptions import NotAuthorized
from unicex.types import NumberLike, RequestMethod
from unicex.utils import filter_params, generate_hmac_sha256_signature


class Client(BaseClient):
    """Клиент для работы с OKX API."""

    _BASE_URL: str = "https://www.okx.com"
    """Базовый URL для REST API OKX."""

    def is_authorized(self) -> bool:
        """Проверяет наличие API‑ключей у клиента.

        Возвращает:
            `bool`: Признак наличия ключей.
        """
        return (
            self._api_key is not None
            and self._api_secret is not None
            and self._api_passphrase is not None
        )

    def _get_timestamp(self) -> str:
        """Генерирует timestamp в формате OKX (ISO с миллисекундами и Z).

        Возвращает:
            `str`: Временная метка в формате ISO с миллисекундами и суффиксом Z.
        """
        now = datetime.datetime.now(tz=datetime.UTC).replace(tzinfo=None)
        timestamp = now.isoformat("T", "milliseconds")
        return timestamp + "Z"

    def _sign_message(
        self,
        method: RequestMethod,
        endpoint: str,
        params: dict[str, Any] | None,
        body: dict[str, Any] | None,
    ) -> tuple[str, str]:
        """Создает timestamp и signature для приватного запроса.

        Алгоритм:
            - формирует строку prehash из timestamp, метода, endpoint, query и body
            - подписывает строку секретным ключом (HMAC-SHA256)
            - кодирует результат в base64

        Параметры:
            method (`RequestMethod`): HTTP-метод (GET, POST и т.д.).
            endpoint (`str`): Относительный путь эндпоинта (например `/api/v5/public/time`).
            params (`dict[str, Any] | None`): Query-параметры.
            body (`dict[str, Any] | None`): Тело запроса (для POST/PUT).

        Возвращает:
            tuple:
                - `timestamp (str)`: Временная метка в формате OKX.
                - `signature (str)`: Подпись в формате base64.
        """
        if not self.is_authorized():
            raise NotAuthorized(
                "Api key and api secret and api passphrase is required to private endpoints"
            )

        timestamp = self._get_timestamp()

        # Формируем query string для GET запросов
        query_string = ""
        if params and method == "GET":
            query_params = "&".join(f"{k}={v}" for k, v in params.items())
            query_string = f"?{query_params}"

        # Формируем body для POST запросов
        body_str = json.dumps(body) if body else ""

        # Создаем строку для подписи: timestamp + method + requestPath + body
        prehash = f"{timestamp}{method}{endpoint}{query_string}{body_str}"
        signature = generate_hmac_sha256_signature(
            self._api_secret,  # type: ignore[arg-type]
            prehash,
            "base64",
        )
        return timestamp, signature

    def _get_headers(self, timestamp: str, signature: str) -> dict[str, str]:
        """Возвращает заголовки для REST-запросов OKX.

        Параметры:
            timestamp (`str`): Временная метка.
            signature (`str`): Подпись (base64).

        Возвращает:
            `dict[str, str]`: Словарь заголовков запроса.
        """
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        headers.update(
            {
                "OK-ACCESS-KEY": self._api_key,  # type: ignore[attr-defined]
                "OK-ACCESS-SIGN": signature,
                "OK-ACCESS-TIMESTAMP": timestamp,
                "OK-ACCESS-PASSPHRASE": self._api_passphrase,  # type: ignore[attr-defined]
                "x-simulated-trading": "0",
            }
        )
        return headers

    def _prepare_request_params(
        self,
        *,
        method: RequestMethod,
        endpoint: str,
        signed: bool,
        params: dict[str, Any] | None,
        body: dict[str, Any] | None = None,
    ) -> tuple[str, dict[str, Any] | None, dict[str, Any] | None, dict[str, str] | None]:
        """Готовит данные для запроса.

        Если signed=True:
            - генерирует timestamp и signature
            - добавляет авторизационные заголовки

        Если signed=False:
            - возвращает только url и переданные параметры.

        Параметры:
            method (`RequestMethod`): HTTP-метод (GET, POST и т.д.).
            endpoint (`str`): Относительный путь эндпоинта.
            signed (`bool`): Нужно ли подписывать запрос.
            params (`dict[str, Any] | None`): Query-параметры.
            body (`dict[str, Any] | None`): Тело запроса.

        Возвращает:
            tuple:
                - `url (str)`: Полный URL для запроса.
                - `params (dict | None)`: Query-параметры.
                - `body (dict | None)`: Тело запроса.
                - `headers (dict | None)`: Заголовки (если signed=True).
        """
        url = f"{self._BASE_URL}{endpoint}"

        # Предобрабатывает параметры запроса
        if params:
            params = filter_params(params)

        headers = None
        if signed:
            timestamp, signature = self._sign_message(method, endpoint, params, body)
            headers = self._get_headers(timestamp, signature)
        return url, params, body, headers

    async def _make_request(
        self,
        method: RequestMethod,
        endpoint: str,
        signed: bool = False,
        *,
        params: dict[str, Any] | None = None,
        data: dict[str, Any] | None = None,
    ) -> Any:
        """Выполняет HTTP-запрос к эндпоинтам OKX API.

        Если `signed=True`:
            - генерирует `timestamp` и `signature`;
            - добавляет авторизационные заголовки (`OK-ACCESS-KEY`, `OK-ACCESS-PASSPHRASE`, `OK-ACCESS-TIMESTAMP`, `OK-ACCESS-SIGN`).

        Если `signed=False`:
            - выполняет публичный запрос без подписи.

        Параметры:
            method (`RequestMethod`): HTTP-метод (`"GET"`, `"POST"`, и т. п.).
            endpoint (`str`): Относительный путь эндпоинта (например, `"/api/v5/public/time"`).
            signed (`bool`): Приватный запрос (с подписью) или публичный. По умолчанию `False`.
            params (`dict[str, Any] | None`): Query-параметры запроса.
            data (`dict[str, Any] | None`): Тело запроса для `POST/PUT`.

        Возвращает:
            `Any`: Ответ API в формате JSON (`dict` или `list`), как вернул сервер.
        """
        url, params, data, headers = self._prepare_request_params(
            method=method,
            endpoint=endpoint,
            signed=signed,
            params=params,
            body=data,
        )
        return await super()._make_request(
            method=method,
            url=url,
            params=params,
            data=data,
            headers=headers,
        )

    async def request(
        self, method: RequestMethod, endpoint: str, params: dict, data: dict, signed: bool
    ) -> dict:
        """Специальный метод для выполнения запросов на эндпоинты, которые не обернуты в клиенте.

        Параметры:
            method (`RequestMethod`): HTTP-метод (`"GET"`, `"POST"`, и т. п.).
            endpoint (`str`): Относительный путь эндпоинта (например, `"/api/v5/public/time"`).
            signed (`bool`): Приватный запрос (с подписью) или публичный.
            params (`dict[str, Any] | None`): Query-параметры запроса.
            data (`dict[str, Any] | None`): Тело запроса для `POST/PUT`.

        Возвращает:
            `dict`: Ответ в формате JSON.
        """
        return await self._make_request(
            method=method, endpoint=endpoint, params=params, data=data, signed=signed
        )

    # topic: Trading Account

    async def get_account_instruments(
        self,
        inst_type: Literal["SPOT", "MARGIN", "SWAP", "FUTURES", "OPTION"],
        inst_family: str | None = None,
        inst_id: str | None = None,
    ) -> dict:
        """Получение доступных инструментов аккаунта.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-instruments
        """
        params = {
            "instType": inst_type,
            "instFamily": inst_family,
            "instId": inst_id,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/instruments",
            params=params,
            signed=True,
        )

    async def get_balance(self, ccy: str | None = None) -> dict:
        """Получение баланса аккаунта.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-balance
        """
        params = {
            "ccy": ccy,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/balance",
            params=params,
            signed=True,
        )

    async def get_positions(
        self,
        inst_type: Literal["MARGIN", "SWAP", "FUTURES", "OPTION"] | None = None,
        inst_id: str | None = None,
        pos_id: str | None = None,
    ) -> dict:
        """Получение текущих позиций аккаунта.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-positions
        """
        params = {
            "instType": inst_type,
            "instId": inst_id,
            "posId": pos_id,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/positions",
            params=params,
            signed=True,
        )

    async def get_positions_history(
        self,
        inst_type: Literal["MARGIN", "SWAP", "FUTURES", "OPTION"] | None = None,
        inst_id: str | None = None,
        mgn_mode: Literal["cross", "isolated"] | None = None,
        type_: str | None = None,
        pos_id: str | None = None,
        after: int | None = None,
        before: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение истории позиций аккаунта.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-positions-history
        """
        params = {
            "instType": inst_type,
            "instId": inst_id,
            "mgnMode": mgn_mode,
            "type": type_,
            "posId": pos_id,
            "after": after,
            "before": before,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/positions-history",
            params=params,
            signed=True,
        )

    async def get_account_position_risk(
        self, inst_type: Literal["MARGIN", "SWAP", "FUTURES", "OPTION"] | None = None
    ) -> dict:
        """Получение риска по аккаунту и позициям.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-account-and-position-risk
        """
        params = {
            "instType": inst_type,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/account-position-risk",
            params=params,
            signed=True,
        )

    async def get_account_bills(
        self,
        inst_type: Literal["SPOT", "MARGIN", "SWAP", "FUTURES", "OPTION"] | None = None,
        inst_id: str | None = None,
        ccy: str | None = None,
        mgn_mode: Literal["isolated", "cross"] | None = None,
        ct_type: Literal["linear", "inverse"] | None = None,
        type_: str | None = None,
        sub_type: str | None = None,
        after: str | None = None,
        before: str | None = None,
        begin: int | None = None,
        end: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение выписок за последние 7 дней.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-bills-details-last-7-days
        """
        params = {
            "instType": inst_type,
            "instId": inst_id,
            "ccy": ccy,
            "mgnMode": mgn_mode,
            "ctType": ct_type,
            "type": type_,
            "subType": sub_type,
            "after": after,
            "before": before,
            "begin": begin,
            "end": end,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/bills",
            params=params,
            signed=True,
        )

    async def get_account_bills_archive(
        self,
        inst_type: Literal["SPOT", "MARGIN", "SWAP", "FUTURES", "OPTION"] | None = None,
        inst_id: str | None = None,
        ccy: str | None = None,
        mgn_mode: Literal["isolated", "cross"] | None = None,
        ct_type: Literal["linear", "inverse"] | None = None,
        type_: str | None = None,
        sub_type: str | None = None,
        after: str | None = None,
        before: str | None = None,
        begin: int | None = None,
        end: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение выписок за последние 3 месяца.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-bills-details-last-3-months
        """
        params = {
            "instType": inst_type,
            "instId": inst_id,
            "ccy": ccy,
            "mgnMode": mgn_mode,
            "ctType": ct_type,
            "type": type_,
            "subType": sub_type,
            "after": after,
            "before": before,
            "begin": begin,
            "end": end,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/bills-archive",
            params=params,
            signed=True,
        )

    async def apply_bills_history_archive(
        self, year: str, quarter: Literal["Q1", "Q2", "Q3", "Q4"]
    ) -> dict:
        """Запрос на формирование архива выписок.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-apply-bills-details-since-2021
        """
        data = {
            "year": year,
            "quarter": quarter,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/bills-history-archive",
            data=data,
            signed=True,
        )

    async def get_bills_history_archive(
        self, year: str, quarter: Literal["Q1", "Q2", "Q3", "Q4"]
    ) -> dict:
        """Получение ссылки на архив выписок.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-bills-details-since-2021
        """
        params = {
            "year": year,
            "quarter": quarter,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/bills-history-archive",
            params=params,
            signed=True,
        )

    async def get_account_config(self) -> dict:
        """Получение конфигурации аккаунта.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-account-configuration
        """
        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/config",
            signed=True,
        )

    async def set_position_mode(self, pos_mode: Literal["long_short_mode", "net_mode"]) -> dict:
        """Изменение режима позиций.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-set-position-mode
        """
        data = {
            "posMode": pos_mode,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/set-position-mode",
            data=data,
            signed=True,
        )

    async def set_leverage(
        self,
        lever: str,
        mgn_mode: Literal["isolated", "cross"],
        inst_id: str | None = None,
        ccy: str | None = None,
        pos_side: Literal["long", "short"] | None = None,
    ) -> dict:
        """Установка плеча.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-set-leverage
        """
        data = {
            "instId": inst_id,
            "ccy": ccy,
            "lever": lever,
            "mgnMode": mgn_mode,
            "posSide": pos_side,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/set-leverage",
            data=data,
            signed=True,
        )

    async def get_max_order_size(
        self,
        inst_id: str,
        td_mode: Literal["cross", "isolated", "cash", "spot_isolated"],
        ccy: str | None = None,
        px: NumberLike | None = None,
        leverage: str | None = None,
        trade_quote_ccy: str | None = None,
    ) -> dict:
        """Получение максимального размера ордера.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-maximum-order-quantity
        """
        params = {
            "instId": inst_id,
            "tdMode": td_mode,
            "ccy": ccy,
            "px": px,
            "leverage": leverage,
            "tradeQuoteCcy": trade_quote_ccy,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/max-size",
            params=params,
            signed=True,
        )

    async def get_max_avail_size(
        self,
        inst_id: str,
        td_mode: Literal["cross", "isolated", "cash", "spot_isolated"],
        ccy: str | None = None,
        reduce_only: bool | None = None,
        px: NumberLike | None = None,
        trade_quote_ccy: str | None = None,
    ) -> dict:
        """Получение максимального доступного баланса/эквити.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-maximum-available-balance-equity
        """
        params = {
            "instId": inst_id,
            "ccy": ccy,
            "tdMode": td_mode,
            "reduceOnly": reduce_only,
            "px": px,
            "tradeQuoteCcy": trade_quote_ccy,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/max-avail-size",
            params=params,
            signed=True,
        )

    async def adjustment_margin(
        self,
        inst_id: str,
        pos_side: Literal["long", "short", "net"],
        type_: Literal["add", "reduce"],
        amt: str,
        ccy: str | None = None,
    ) -> dict:
        """Изменение маржи изолированной позиции.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-increase-decrease-margin
        """
        data = {
            "instId": inst_id,
            "posSide": pos_side,
            "type": type_,
            "amt": amt,
            "ccy": ccy,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/position/margin-balance",
            data=data,
            signed=True,
        )

    async def get_leverage(
        self,
        mgn_mode: Literal["isolated", "cross"],
        inst_id: str | None = None,
        ccy: str | None = None,
    ) -> dict:
        """Получение текущего плеча.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-leverage
        """
        params = {
            "instId": inst_id,
            "ccy": ccy,
            "mgnMode": mgn_mode,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/leverage-info",
            params=params,
            signed=True,
        )

    async def get_leverage_estimated_info(
        self,
        inst_type: Literal["MARGIN", "SWAP", "FUTURES"],
        mgn_mode: Literal["isolated", "cross"],
        lever: str,
        inst_id: str | None = None,
        ccy: str | None = None,
        pos_side: Literal["net", "long", "short"] | None = None,
    ) -> dict:
        """Получение оценочных данных по плечу.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-leverage-estimated-info
        """
        params = {
            "instType": inst_type,
            "mgnMode": mgn_mode,
            "lever": lever,
            "instId": inst_id,
            "ccy": ccy,
            "posSide": pos_side,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/adjust-leverage-info",
            params=params,
            signed=True,
        )

    async def get_max_loan(
        self,
        mgn_mode: Literal["isolated", "cross"],
        inst_id: str | None = None,
        ccy: str | None = None,
        mgn_ccy: str | None = None,
        trade_quote_ccy: str | None = None,
    ) -> dict:
        """Получение максимального объема заимствования.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-the-maximum-loan-of-instrument
        """
        params = {
            "mgnMode": mgn_mode,
            "instId": inst_id,
            "ccy": ccy,
            "mgnCcy": mgn_ccy,
            "tradeQuoteCcy": trade_quote_ccy,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/max-loan",
            params=params,
            signed=True,
        )

    async def get_fee_rates(
        self,
        inst_type: Literal["SPOT", "MARGIN", "SWAP", "FUTURES", "OPTION"],
        inst_id: str | None = None,
        inst_family: str | None = None,
        rule_type: Literal["normal", "pre_market"] | None = None,
    ) -> dict:
        """Получение торговых комиссий.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-fee-rates
        """
        params = {
            "instType": inst_type,
            "instId": inst_id,
            "instFamily": inst_family,
            "ruleType": rule_type,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/trade-fee",
            params=params,
            signed=True,
        )

    async def get_interest_accrued(
        self,
        type_: str | None = None,
        ccy: str | None = None,
        inst_id: str | None = None,
        mgn_mode: Literal["cross", "isolated"] | None = None,
        after: int | None = None,
        before: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение данных по начисленным процентам.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-interest-accrued-data
        """
        params = {
            "type": type_,
            "ccy": ccy,
            "instId": inst_id,
            "mgnMode": mgn_mode,
            "after": after,
            "before": before,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/interest-accrued",
            params=params,
            signed=True,
        )

    async def get_interest_rate(self, ccy: str | None = None) -> dict:
        """Получение текущих процентных ставок.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-interest-rate
        """
        params = {
            "ccy": ccy,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/interest-rate",
            params=params,
            signed=True,
        )

    async def set_fee_type(self, fee_type: Literal["0", "1"]) -> dict:
        """Настройка типа комиссии.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-set-fee-type
        """
        data = {
            "feeType": fee_type,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/set-fee-type",
            data=data,
            signed=True,
        )

    async def set_greeks(self, greeks_type: Literal["PA", "BS"]) -> dict:
        """Настройка формата греков.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-set-greeks-pa-bs
        """
        data = {
            "greeksType": greeks_type,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/set-greeks",
            data=data,
            signed=True,
        )

    async def set_isolated_mode(
        self,
        iso_mode: Literal["auto_transfers_ccy", "automatic"],
        type_: Literal["MARGIN", "CONTRACTS"],
    ) -> dict:
        """Настройка режима изолированной маржи.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-isolated-margin-trading-settings
        """
        data = {
            "isoMode": iso_mode,
            "type": type_,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/set-isolated-mode",
            data=data,
            signed=True,
        )

    async def get_max_withdrawal(self, ccy: str | None = None) -> dict:
        """Получение максимальной суммы вывода из трейдингового аккаунта.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-maximum-withdrawals
        """
        params = {
            "ccy": ccy,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/max-withdrawal",
            params=params,
            signed=True,
        )

    async def get_risk_state(self) -> dict:
        """Получение статуса рисков аккаунта.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-account-risk-state
        """
        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/risk-state",
            signed=True,
        )

    async def get_interest_limits(
        self,
        type_: str | None = None,
        ccy: str | None = None,
    ) -> dict:
        """Получение лимитов и процентов заимствования.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-borrow-interest-and-limit
        """
        params = {
            "type": type_,
            "ccy": ccy,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/interest-limits",
            params=params,
            signed=True,
        )

    async def spot_manual_borrow_repay(
        self, ccy: str, side: Literal["borrow", "repay"], amt: str
    ) -> dict:
        """Ручное заимствование или погашение в спотовом режиме.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-manual-borrow-repay
        """
        data = {
            "ccy": ccy,
            "side": side,
            "amt": amt,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/spot-manual-borrow-repay",
            data=data,
            signed=True,
        )

    async def set_auto_repay(self, auto_repay: bool) -> dict:
        """Настройка автоматического погашения в спотовом режиме.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-set-auto-repay
        """
        data = {
            "autoRepay": auto_repay,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/set-auto-repay",
            data=data,
            signed=True,
        )

    async def spot_borrow_repay_history(
        self,
        ccy: str | None = None,
        type_: Literal["auto_borrow", "auto_repay", "manual_borrow", "manual_repay"] | None = None,
        after: int | None = None,
        before: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение истории заимствований и погашений в спотовом режиме.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-borrow-repay-history
        """
        params = {
            "ccy": ccy,
            "type": type_,
            "after": after,
            "before": before,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/spot-borrow-repay-history",
            params=params,
            signed=True,
        )

    async def position_builder(
        self,
        acct_lv: str | None = None,
        incl_real_pos_and_eq: bool | None = None,
        lever: str | None = None,
        sim_pos: list[dict[str, Any]] | None = None,
        sim_asset: list[dict[str, Any]] | None = None,
        greeks_type: Literal["BS", "PA", "CASH"] | None = None,
        idx_vol: str | None = None,
    ) -> dict:
        """Расчет параметров портфеля с виртуальными позициями.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-position-builder-new
        """
        data = {
            "acctLv": acct_lv,
            "inclRealPosAndEq": incl_real_pos_and_eq,
            "lever": lever,
            "simPos": sim_pos,
            "simAsset": sim_asset,
            "greeksType": greeks_type,
            "idxVol": idx_vol,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/position-builder",
            data=data,
            signed=True,
        )

    async def position_builder_graph(
        self,
        type_: Literal["mmr"],
        incl_real_pos_and_eq: bool | None = None,
        sim_pos: list[dict[str, Any]] | None = None,
        sim_asset: list[dict[str, Any]] | None = None,
        greeks_type: Literal["BS", "PA", "CASH"] | None = None,
        mmr_config: dict[str, Any] | None = None,
    ) -> dict:
        """Расчет графика тренда риск-параметров.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-position-builder-trend-graph
        """
        data = {
            "inclRealPosAndEq": incl_real_pos_and_eq,
            "simPos": sim_pos,
            "simAsset": sim_asset,
            "greeksType": greeks_type,
            "type": type_,
            "mmrConfig": mmr_config,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/position-builder-graph",
            data=data,
            signed=True,
        )

    async def set_risk_offset_amount(self, ccy: str, cl_spot_in_use_amt: str) -> dict:
        """Установка величины спотового риск-офсета.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-set-risk-offset-amount
        """
        data = {
            "ccy": ccy,
            "clSpotInUseAmt": cl_spot_in_use_amt,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/set-riskOffset-amt",
            data=data,
            signed=True,
        )

    async def get_greeks(self, ccy: str | None = None) -> dict:
        """Получение греков по активам.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-greeks
        """
        params = {
            "ccy": ccy,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/greeks",
            params=params,
            signed=True,
        )

    async def get_account_position_tiers(
        self,
        inst_type: Literal["SWAP", "FUTURES", "OPTION"],
        inst_family: str,
        uly: str | None = None,
    ) -> dict:
        """Получение лимитов позиций в PM режиме.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-pm-position-limitation
        """
        params = {
            "instType": inst_type,
            "instFamily": inst_family,
            "uly": uly,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/position-tiers",
            params=params,
            signed=True,
        )

    async def activate_option(self) -> dict:
        """Активация торговли опционами.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-activate-option
        """
        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/activate-option",
            data={},
            signed=True,
        )

    async def set_auto_loan(self, auto_loan: bool | None = None) -> dict:
        """Настройка автоматического заимствования.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-set-auto-loan
        """
        data = {
            "autoLoan": auto_loan,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/set-auto-loan",
            data=data,
            signed=True,
        )

    async def account_level_switch_preset(
        self, acct_lv: Literal["2", "3", "4"], lever: str | None = None
    ) -> dict:
        """Преднастройка смены режима аккаунта.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-preset-account-mode-switch
        """
        data = {
            "acctLv": acct_lv,
            "lever": lever,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/account-level-switch-preset",
            data=data,
            signed=True,
        )

    async def get_account_switch_precheck(self, acct_lv: Literal["1", "2", "3", "4"]) -> dict:
        """Предварительная проверка смены режима аккаунта.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-precheck-account-mode-switch
        """
        params = {
            "acctLv": acct_lv,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/set-account-switch-precheck",
            params=params,
            signed=True,
        )

    async def set_account_level(self, acct_lv: Literal["1", "2", "3", "4"]) -> dict:
        """Смена режима аккаунта.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-set-account-mode
        """
        data = {
            "acctLv": acct_lv,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/set-account-level",
            data=data,
            signed=True,
        )

    async def set_collateral_assets(
        self,
        type_: Literal["all", "custom"],
        collateral_enabled: bool,
        ccy_list: list[str] | None = None,
    ) -> dict:
        """Настройка списка коллатеральных активов.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-set-collateral-assets
        """
        data = {
            "type": type_,
            "collateralEnabled": collateral_enabled,
            "ccyList": ccy_list,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/set-collateral-assets",
            data=data,
            signed=True,
        )

    async def get_collateral_assets(
        self, ccy: str | None = None, collateral_enabled: bool | None = None
    ) -> dict:
        """Получение списка коллатеральных активов.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-collateral-assets
        """
        params = {
            "ccy": ccy,
            "collateralEnabled": collateral_enabled,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/collateral-assets",
            params=params,
            signed=True,
        )

    async def reset_mmp_status(
        self, inst_family: str, inst_type: Literal["OPTION"] | None = None
    ) -> dict:
        """Сброс статуса MMP.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-reset-mmp-status
        """
        data = {
            "instType": inst_type,
            "instFamily": inst_family,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/mmp-reset",
            data=data,
            signed=True,
        )

    async def set_mmp(
        self,
        inst_family: str,
        time_interval: str,
        frozen_interval: str,
        qty_limit: str,
    ) -> dict:
        """Настройка параметров MMP.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-set-mmp
        """
        data = {
            "instFamily": inst_family,
            "timeInterval": time_interval,
            "frozenInterval": frozen_interval,
            "qtyLimit": qty_limit,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/mmp-config",
            data=data,
            signed=True,
        )

    async def get_mmp_config(self, inst_family: str | None = None) -> dict:
        """Получение настроек MMP.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-mmp-config
        """
        params = {
            "instFamily": inst_family,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/mmp-config",
            params=params,
            signed=True,
        )

    async def move_positions(
        self,
        from_acct: str,
        to_acct: str,
        legs: list[dict[str, Any]],
        client_id: str,
    ) -> dict:
        """Перемещение позиций между аккаунтами.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-move-positions
        """
        data = {
            "fromAcct": from_acct,
            "toAcct": to_acct,
            "legs": legs,
            "clientId": client_id,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/move-positions",
            data=data,
            signed=True,
        )

    async def get_move_positions_history(
        self,
        block_td_id: str | None = None,
        client_id: str | None = None,
        begin_ts: int | None = None,
        end_ts: int | None = None,
        limit: int | None = None,
        state: Literal["filled", "pending"] | None = None,
    ) -> dict:
        """Получение истории перемещения позиций.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-move-positions-history
        """
        params = {
            "blockTdId": block_td_id,
            "clientId": client_id,
            "beginTs": begin_ts,
            "endTs": end_ts,
            "limit": limit,
            "state": state,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/account/move-positions-history",
            params=params,
            signed=True,
        )

    async def set_auto_earn(
        self,
        ccy: str,
        action: Literal["turn_on", "turn_off"],
        earn_type: Literal["0", "1"] | None = None,
    ) -> dict:
        """Настройка автоматического Earn.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-set-auto-earn
        """
        data = {
            "earnType": earn_type,
            "ccy": ccy,
            "action": action,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/set-auto-earn",
            data=data,
            signed=True,
        )

    async def set_settle_currency(self, settle_ccy: str) -> dict:
        """Настройка расчетной валюты для USD-маржинальных контрактов.

        https://www.okx.com/docs-v5/en/#trading-account-rest-api-set-settle-currency
        """
        data = {
            "settleCcy": settle_ccy,
        }

        return await self._make_request(
            "POST",
            endpoint="/api/v5/account/set-settle-currency",
            data=data,
            signed=True,
        )

    # topic: Order Book Trading (Trade)

    async def place_order(
        self,
        inst_id: str,
        td_mode: Literal["cross", "isolated", "cash", "spot_isolated"],
        side: Literal["buy", "sell"],
        ord_type: Literal[
            "market",
            "limit",
            "post_only",
            "fok",
            "ioc",
            "optimal_limit_ioc",
            "mmp",
            "mmp_and_post_only",
            "op_fok",
        ],
        sz: NumberLike,
        ccy: str | None = None,
        cl_ord_id: str | None = None,
        tag: str | None = None,
        pos_side: Literal["net", "long", "short"] | None = None,
        px: NumberLike | None = None,
        px_usd: NumberLike | None = None,
        px_vol: NumberLike | None = None,
        reduce_only: bool | None = None,
        tgt_ccy: Literal["base_ccy", "quote_ccy"] | None = None,
        ban_amend: bool | None = None,
        px_amend_type: Literal["0", "1"] | None = None,
        trade_quote_ccy: str | None = None,
        stp_mode: Literal["cancel_maker", "cancel_taker", "cancel_both"] | None = None,
        quick_mgn_type: str | None = None,
        attach_algo_orders: list[dict[str, Any]] | None = None,
        extra_params: dict[str, Any] | None = None,
    ) -> dict:
        """Создание ордера.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-place-order
        """
        data: dict[str, Any] = {
            "instId": inst_id,
            "tdMode": td_mode,
            "side": side,
            "ordType": ord_type,
            "sz": sz,
            "ccy": ccy,
            "clOrdId": cl_ord_id,
            "tag": tag,
            "posSide": pos_side,
            "px": px,
            "pxUsd": px_usd,
            "pxVol": px_vol,
            "reduceOnly": reduce_only,
            "tgtCcy": tgt_ccy,
            "banAmend": ban_amend,
            "pxAmendType": px_amend_type,
            "tradeQuoteCcy": trade_quote_ccy,
            "stpMode": stp_mode,
            "quickMgnType": quick_mgn_type,
        }
        if attach_algo_orders is not None:
            data["attachAlgoOrds"] = [filter_params(order) for order in attach_algo_orders]
        if extra_params:
            data.update(extra_params)

        return await self._make_request(
            "POST",
            endpoint="/api/v5/trade/order",
            data=filter_params(data),
            signed=True,
        )

    async def place_multiple_orders(self, orders: list[dict[str, Any]]) -> dict:
        """Создание ордеров пакетно.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-place-multiple-orders
        """
        data = [filter_params(order) for order in orders]

        return await self._make_request(
            "POST",
            endpoint="/api/v5/trade/batch-orders",
            data=data,  # type: ignore
            signed=True,
        )

    async def cancel_order(
        self,
        inst_id: str,
        ord_id: str | None = None,
        cl_ord_id: str | None = None,
    ) -> dict:
        """Отмена ордера.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-cancel-order
        """
        if not ord_id and not cl_ord_id:
            raise ValueError("Either ord_id or cl_ord_id must be provided")

        data = filter_params(
            {
                "instId": inst_id,
                "ordId": ord_id,
                "clOrdId": cl_ord_id,
            }
        )

        return await self._make_request(
            "POST",
            endpoint="/api/v5/trade/cancel-order",
            data=data,
            signed=True,
        )

    async def cancel_multiple_orders(self, orders: list[dict[str, Any]]) -> dict:
        """Отмена ордеров пакетно.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-cancel-multiple-orders
        """
        data = [filter_params(order) for order in orders]

        return await self._make_request(
            "POST",
            endpoint="/api/v5/trade/cancel-batch-orders",
            data=data,  # type: ignore
            signed=True,
        )

    async def amend_order(
        self,
        inst_id: str,
        ord_id: str | None = None,
        cl_ord_id: str | None = None,
        *,
        new_sz: NumberLike | None = None,
        new_px: NumberLike | None = None,
        new_px_usd: NumberLike | None = None,
        new_px_vol: NumberLike | None = None,
        cxl_on_fail: bool | None = None,
        req_id: str | None = None,
        px_amend_type: Literal["0", "1"] | None = None,
        attach_algo_orders: list[dict[str, Any]] | None = None,
        extra_params: dict[str, Any] | None = None,
    ) -> dict:
        """Изменение параметров ордера.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-amend-order
        """
        if not ord_id and not cl_ord_id:
            raise ValueError("Either ord_id or cl_ord_id must be provided")

        data: dict[str, Any] = {
            "instId": inst_id,
            "ordId": ord_id,
            "clOrdId": cl_ord_id,
            "newSz": new_sz,
            "newPx": new_px,
            "newPxUsd": new_px_usd,
            "newPxVol": new_px_vol,
            "cxlOnFail": cxl_on_fail,
            "reqId": req_id,
            "pxAmendType": px_amend_type,
        }
        if attach_algo_orders is not None:
            data["attachAlgoOrds"] = [filter_params(order) for order in attach_algo_orders]
        if extra_params:
            data.update(extra_params)

        return await self._make_request(
            "POST",
            endpoint="/api/v5/trade/amend-order",
            data=filter_params(data),
            signed=True,
        )

    async def amend_multiple_orders(self, orders: list[dict[str, Any]]) -> dict:
        """Изменение параметров ордеров пакетно.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-amend-multiple-orders
        """
        data = [filter_params(order) for order in orders]

        return await self._make_request(
            "POST",
            endpoint="/api/v5/trade/amend-batch-orders",
            data=data,  # type: ignore
            signed=True,
        )

    async def close_positions(
        self,
        inst_id: str,
        mgn_mode: Literal["cross", "isolated"],
        pos_side: Literal["net", "long", "short"] | None = None,
        ccy: str | None = None,
        auto_cxl: bool | None = None,
        cl_ord_id: str | None = None,
        tag: str | None = None,
        extra_params: dict[str, Any] | None = None,
    ) -> dict:
        """Закрытие позиции рыночным ордером.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-close-positions
        """
        data: dict[str, Any] = {
            "instId": inst_id,
            "mgnMode": mgn_mode,
            "posSide": pos_side,
            "ccy": ccy,
            "autoCxl": auto_cxl,
            "clOrdId": cl_ord_id,
            "tag": tag,
        }
        if extra_params:
            data.update(extra_params)

        return await self._make_request(
            "POST",
            endpoint="/api/v5/trade/close-position",
            data=filter_params(data),
            signed=True,
        )

    async def get_order(
        self,
        inst_id: str,
        ord_id: str | None = None,
        cl_ord_id: str | None = None,
    ) -> dict:
        """Получение информации об ордере.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-get-order-details
        """
        if not ord_id and not cl_ord_id:
            raise ValueError("Either ord_id or cl_ord_id must be provided")

        params = filter_params(
            {
                "instId": inst_id,
                "ordId": ord_id,
                "clOrdId": cl_ord_id,
            }
        )

        return await self._make_request(
            "GET",
            endpoint="/api/v5/trade/order",
            params=params,
            signed=True,
        )

    async def get_order_list(
        self,
        inst_type: str | None = None,
        inst_family: str | None = None,
        inst_id: str | None = None,
        ord_type: str | None = None,
        state: Literal["live", "partially_filled"] | None = None,
        after: str | None = None,
        before: str | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение списка активных ордеров.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-get-order-list
        """
        params = filter_params(
            {
                "instType": inst_type,
                "instFamily": inst_family,
                "instId": inst_id,
                "ordType": ord_type,
                "state": state,
                "after": after,
                "before": before,
                "limit": limit,
            }
        )

        return await self._make_request(
            "GET",
            endpoint="/api/v5/trade/orders-pending",
            params=params,
            signed=True,
        )

    async def get_orders_history(
        self,
        inst_type: str,
        inst_family: str | None = None,
        inst_id: str | None = None,
        ord_type: str | None = None,
        state: str | None = None,
        category: str | None = None,
        after: str | None = None,
        before: str | None = None,
        begin: int | None = None,
        end: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение истории ордеров за 7 дней.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-get-order-history-last-7-days
        """
        params = filter_params(
            {
                "instType": inst_type,
                "instFamily": inst_family,
                "instId": inst_id,
                "ordType": ord_type,
                "state": state,
                "category": category,
                "after": after,
                "before": before,
                "begin": begin,
                "end": end,
                "limit": limit,
            }
        )

        return await self._make_request(
            "GET",
            endpoint="/api/v5/trade/orders-history",
            params=params,
            signed=True,
        )

    async def get_orders_history_archive(
        self,
        inst_type: str,
        inst_family: str | None = None,
        inst_id: str | None = None,
        ord_type: str | None = None,
        state: str | None = None,
        category: str | None = None,
        after: str | None = None,
        before: str | None = None,
        begin: int | None = None,
        end: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение истории ордеров за 3 месяца.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-get-order-history-last-3-months
        """
        params = filter_params(
            {
                "instType": inst_type,
                "instFamily": inst_family,
                "instId": inst_id,
                "ordType": ord_type,
                "state": state,
                "category": category,
                "after": after,
                "before": before,
                "begin": begin,
                "end": end,
                "limit": limit,
            }
        )

        return await self._make_request(
            "GET",
            endpoint="/api/v5/trade/orders-history-archive",
            params=params,
            signed=True,
        )

    async def get_fills(
        self,
        inst_type: str | None = None,
        inst_family: str | None = None,
        inst_id: str | None = None,
        ord_id: str | None = None,
        sub_type: str | None = None,
        after: str | None = None,
        before: str | None = None,
        begin: int | None = None,
        end: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение сделок за 3 дня.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-get-transaction-details-last-3-days
        """
        params = filter_params(
            {
                "instType": inst_type,
                "instFamily": inst_family,
                "instId": inst_id,
                "ordId": ord_id,
                "subType": sub_type,
                "after": after,
                "before": before,
                "begin": begin,
                "end": end,
                "limit": limit,
            }
        )

        return await self._make_request(
            "GET",
            endpoint="/api/v5/trade/fills",
            params=params,
            signed=True,
        )

    async def get_fills_history(
        self,
        inst_type: str,
        inst_family: str | None = None,
        inst_id: str | None = None,
        ord_id: str | None = None,
        sub_type: str | None = None,
        after: str | None = None,
        before: str | None = None,
        begin: int | None = None,
        end: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение сделок за 3 месяца.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-get-transaction-details-last-3-months
        """
        params = filter_params(
            {
                "instType": inst_type,
                "instFamily": inst_family,
                "instId": inst_id,
                "ordId": ord_id,
                "subType": sub_type,
                "after": after,
                "before": before,
                "begin": begin,
                "end": end,
                "limit": limit,
            }
        )

        return await self._make_request(
            "GET",
            endpoint="/api/v5/trade/fills-history",
            params=params,
            signed=True,
        )

    async def get_easy_convert_currency_list(self, source: Literal["1", "2"] | None = None) -> dict:
        """Получение списка валют для Easy Convert.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-get-easy-convert-currency-list
        """
        params = filter_params({"source": source})

        return await self._make_request(
            "GET",
            endpoint="/api/v5/trade/easy-convert-currency-list",
            params=params,
            signed=True,
        )

    async def easy_convert(
        self,
        from_ccy: list[str],
        to_ccy: str,
        source: Literal["1", "2"] | None = None,
    ) -> dict:
        """Конвертация мелких остатков через Easy Convert.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-place-easy-convert
        """
        data = filter_params(
            {
                "fromCcy": from_ccy,
                "toCcy": to_ccy,
                "source": source,
            }
        )

        return await self._make_request(
            "POST",
            endpoint="/api/v5/trade/easy-convert",
            data=data,
            signed=True,
        )

    async def get_easy_convert_history(
        self,
        after: str | None = None,
        before: str | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение истории Easy Convert.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-get-easy-convert-history
        """
        params = filter_params(
            {
                "after": after,
                "before": before,
                "limit": limit,
            }
        )

        return await self._make_request(
            "GET",
            endpoint="/api/v5/trade/easy-convert-history",
            params=params,
            signed=True,
        )

    async def get_one_click_repay_currency_list(
        self,
        debt_type: Literal["cross", "isolated"] | None = None,
    ) -> dict:
        """Получение списка валют для One-click Repay.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-get-one-click-repay-currency-list
        """
        params = filter_params({"debtType": debt_type})

        return await self._make_request(
            "GET",
            endpoint="/api/v5/trade/one-click-repay-currency-list",
            params=params,
            signed=True,
        )

    async def trade_one_click_repay(
        self,
        debt_ccy: list[str],
        repay_ccy: str,
    ) -> dict:
        """Совершение One-click Repay.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-trade-one-click-repay
        """
        data = filter_params({"debtCcy": debt_ccy, "repayCcy": repay_ccy})

        return await self._make_request(
            "POST",
            endpoint="/api/v5/trade/one-click-repay",
            data=data,
            signed=True,
        )

    async def get_one_click_repay_history(
        self,
        after: str | None = None,
        before: str | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение истории One-click Repay.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-get-one-click-repay-history
        """
        params = filter_params(
            {
                "after": after,
                "before": before,
                "limit": limit,
            }
        )

        return await self._make_request(
            "GET",
            endpoint="/api/v5/trade/one-click-repay-history",
            params=params,
            signed=True,
        )

    async def get_one_click_repay_currency_list_v2(self) -> dict:
        """Получение списка валют для One-click Repay (v2).

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-get-one-click-repay-currency-list-new
        """
        return await self._make_request(
            "GET",
            endpoint="/api/v5/trade/one-click-repay-currency-list-v2",
            signed=True,
        )

    async def trade_one_click_repay_v2(
        self,
        debt_ccy: str,
        repay_ccy_list: list[str],
    ) -> dict:
        """Совершение One-click Repay (v2).

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-trade-one-click-repay-new
        """
        data = filter_params({"debtCcy": debt_ccy, "repayCcyList": repay_ccy_list})

        return await self._make_request(
            "POST",
            endpoint="/api/v5/trade/one-click-repay-v2",
            data=data,
            signed=True,
        )

    async def get_one_click_repay_history_v2(
        self,
        after: str | None = None,
        before: str | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение истории One-click Repay (v2).

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-get-one-click-repay-history-new
        """
        params = filter_params(
            {
                "after": after,
                "before": before,
                "limit": limit,
            }
        )

        return await self._make_request(
            "GET",
            endpoint="/api/v5/trade/one-click-repay-history-v2",
            params=params,
            signed=True,
        )

    async def mass_cancel_orders(
        self,
        inst_type: Literal["OPTION"],
        inst_family: str,
        lock_interval: str | None = None,
    ) -> dict:
        """Массовое снятие MMP-ордеров.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-mass-cancel-order
        """
        data = filter_params(
            {
                "instType": inst_type,
                "instFamily": inst_family,
                "lockInterval": lock_interval,
            }
        )

        return await self._make_request(
            "POST",
            endpoint="/api/v5/trade/mass-cancel",
            data=data,
            signed=True,
        )

    async def cancel_all_after(
        self,
        time_out: str,
        tag: str | None = None,
    ) -> dict:
        """Установка Cancel All After.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-cancel-all-after
        """
        data = filter_params({"timeOut": time_out, "tag": tag})

        return await self._make_request(
            "POST",
            endpoint="/api/v5/trade/cancel-all-after",
            data=data,
            signed=True,
        )

    async def get_account_rate_limit(self) -> dict:
        """Получение информации о лимитах запросов аккаунта.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-get-account-rate-limit
        """
        return await self._make_request(
            "GET",
            endpoint="/api/v5/trade/account-rate-limit",
            signed=True,
        )

    async def order_precheck(
        self,
        inst_id: str,
        td_mode: Literal["cross", "isolated", "cash", "spot_isolated"],
        side: Literal["buy", "sell"],
        ord_type: Literal[
            "market",
            "limit",
            "post_only",
            "fok",
            "ioc",
            "optimal_limit_ioc",
        ],
        sz: NumberLike,
        pos_side: Literal["net", "long", "short"] | None = None,
        px: NumberLike | None = None,
        reduce_only: bool | None = None,
        tgt_ccy: Literal["base_ccy", "quote_ccy"] | None = None,
        attach_algo_orders: list[dict[str, Any]] | None = None,
        extra_params: dict[str, Any] | None = None,
    ) -> dict:
        """Превентивная проверка перед размещением ордера.

        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-order-precheck
        """
        data: dict[str, Any] = {
            "instId": inst_id,
            "tdMode": td_mode,
            "side": side,
            "ordType": ord_type,
            "sz": sz,
            "posSide": pos_side,
            "px": px,
            "reduceOnly": reduce_only,
            "tgtCcy": tgt_ccy,
        }
        if attach_algo_orders is not None:
            data["attachAlgoOrds"] = [filter_params(order) for order in attach_algo_orders]
        if extra_params:
            data.update(extra_params)

        return await self._make_request(
            "POST",
            endpoint="/api/v5/trade/order-precheck",
            data=filter_params(data),
            signed=True,
        )

    # topic: Order Book Trading (Market Data)

    async def get_tickers(
        self,
        inst_type: Literal["SPOT", "SWAP", "FUTURES", "OPTION"],
        inst_family: str | None = None,
    ) -> dict:
        """Получение списка тикеров с основными метриками.

        https://www.okx.com/docs-v5/en/#order-book-trading-market-data-get-tickers
        """
        params = {
            "instType": inst_type,
            "instFamily": inst_family,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/market/tickers",
            params=params,
        )

    async def get_ticker(self, inst_id: str) -> dict:
        """Получение тикера инструмента с данными за 24 часа.

        https://www.okx.com/docs-v5/en/#order-book-trading-market-data-get-ticker
        """
        params = {
            "instId": inst_id,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/market/ticker",
            params=params,
        )

    async def get_order_book(self, inst_id: str, sz: int | None = None) -> dict:
        """Получение книги ордеров с обновлением каждые 50 мс.

        https://www.okx.com/docs-v5/en/#order-book-trading-market-data-get-order-book
        """
        params = {
            "instId": inst_id,
            "sz": sz,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/market/books",
            params=params,
        )

    async def get_full_order_book(self, inst_id: str, sz: int | None = None) -> dict:
        """Получение полной книги ордеров с глубиной до 5000 уровней.

        https://www.okx.com/docs-v5/en/#order-book-trading-market-data-get-full-order-book
        """
        params = {
            "instId": inst_id,
            "sz": sz,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/market/books-full",
            params=params,
        )

    async def get_candlesticks(
        self,
        inst_id: str,
        bar: str | None = None,
        after: int | None = None,
        before: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение списка свечей с максимальной глубиной в 1440 записей.

        https://www.okx.com/docs-v5/en/#order-book-trading-market-data-get-candlesticks
        """
        params = {
            "instId": inst_id,
            "bar": bar,
            "after": after,
            "before": before,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/market/candles",
            params=params,
        )

    async def get_candlesticks_history(
        self,
        inst_id: str,
        after: int | None = None,
        before: int | None = None,
        bar: str | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение исторических свечей за прошлые периоды.

        https://www.okx.com/docs-v5/en/#order-book-trading-market-data-get-candlesticks-history
        """
        params = {
            "instId": inst_id,
            "after": after,
            "before": before,
            "bar": bar,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/market/history-candles",
            params=params,
        )

    async def get_trades(self, inst_id: str, limit: int | None = None) -> dict:
        """Получение последних сделок по инструменту.

        https://www.okx.com/docs-v5/en/#order-book-trading-market-data-get-trades
        """
        params = {
            "instId": inst_id,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/market/trades",
            params=params,
        )

    async def get_trades_history(
        self,
        inst_id: str,
        type_: Literal["1", "2"] | None = None,
        after: str | None = None,
        before: str | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение истории сделок с пагинацией за последние три месяца.

        https://www.okx.com/docs-v5/en/#order-book-trading-market-data-get-trades-history
        """
        params = {
            "instId": inst_id,
            "type": type_,
            "after": after,
            "before": before,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/market/history-trades",
            params=params,
        )

    async def get_option_trades_by_family(self, inst_family: str) -> dict:
        """Получение сделок по всем опционам в рамках семьи инструментов.

        https://www.okx.com/docs-v5/en/#order-book-trading-market-data-get-option-trades-by-instrument-family
        """
        params = {
            "instFamily": inst_family,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/market/option/instrument-family-trades",
            params=params,
        )

    async def get_option_trades(
        self,
        inst_id: str | None = None,
        inst_family: str | None = None,
        opt_type: Literal["C", "P"] | None = None,
    ) -> dict:
        """Получение сделок по выбранным опционам.

        https://www.okx.com/docs-v5/en/#order-book-trading-market-data-get-option-trades
        """
        params = {
            "instId": inst_id,
            "instFamily": inst_family,
            "optType": opt_type,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/option-trades",
            params=params,
        )

    async def get_24h_total_volume(self) -> dict:
        """Получение совокупного 24-часового объема торгов по платформе.

        https://www.okx.com/docs-v5/en/#order-book-trading-market-data-get-24h-total-volume
        """
        return await self._make_request(
            "GET",
            endpoint="/api/v5/market/platform-24-volume",
        )

    async def get_call_auction_details(self, inst_id: str) -> dict:
        """Получение данных по предторговому аукциону инструмента.

        https://www.okx.com/docs-v5/en/#order-book-trading-market-data-get-call-auction-details
        """
        params = {
            "instId": inst_id,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/market/call-auction-details",
            params=params,
        )

    # topic: Public Data

    async def get_instruments(
        self,
        inst_type: Literal["SPOT", "MARGIN", "SWAP", "FUTURES", "OPTION"],
        inst_family: str | None = None,
        inst_id: str | None = None,
    ) -> dict:
        """Получение списка доступных публичных инструментов.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-instruments
        """
        params = {
            "instType": inst_type,
            "instFamily": inst_family,
            "instId": inst_id,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/instruments",
            params=params,
        )

    async def get_estimated_delivery_price(self, inst_id: str) -> dict:
        """Получение оценочной цены поставки или исполнения опциона.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-estimated-delivery-exercise-price
        """
        params = {
            "instId": inst_id,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/estimated-price",
            params=params,
        )

    async def get_delivery_exercise_history(
        self,
        inst_type: Literal["FUTURES", "OPTION"],
        inst_family: str,
        after: int | None = None,
        before: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение истории поставок и исполнений за последние три месяца.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-delivery-exercise-history
        """
        params = {
            "instType": inst_type,
            "instFamily": inst_family,
            "after": after,
            "before": before,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/delivery-exercise-history",
            params=params,
        )

    async def get_estimated_settlement_info(self, inst_id: str) -> dict:
        """Получение оценочной цены ближайшего расчета по фьючерсу.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-estimated-future-settlement-price
        """
        params = {
            "instId": inst_id,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/estimated-settlement-info",
            params=params,
        )

    async def get_futures_settlement_history(
        self,
        inst_family: str,
        after: int | None = None,
        before: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение истории расчетов фьючерсов за последние три месяца.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-futures-settlement-history
        """
        params = {
            "instFamily": inst_family,
            "after": after,
            "before": before,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/settlement-history",
            params=params,
        )

    async def get_funding_rate(self, inst_id: str) -> dict:
        """Получение текущей ставки финансирования.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-funding-rate
        """
        params = {
            "instId": inst_id,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/funding-rate",
            params=params,
        )

    async def get_funding_rate_history(
        self,
        inst_id: str,
        before: int | None = None,
        after: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение истории ставок финансирования за последние три месяца.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-funding-rate-history
        """
        params = {
            "instId": inst_id,
            "before": before,
            "after": after,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/funding-rate-history",
            params=params,
        )

    async def get_open_interest(
        self,
        inst_type: Literal["SWAP", "FUTURES", "OPTION"],
        inst_family: str | None = None,
        inst_id: str | None = None,
    ) -> dict:
        """Получение общего открытого интереса по контрактам OKX.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-open-interest
        """
        params = {
            "instType": inst_type,
            "instFamily": inst_family,
            "instId": inst_id,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/open-interest",
            params=params,
        )

    async def get_price_limit(self, inst_id: str) -> dict:
        """Получение верхнего и нижнего лимитов цен для инструмента.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-limit-price
        """
        params = {
            "instId": inst_id,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/price-limit",
            params=params,
        )

    async def get_option_market_data(
        self,
        inst_family: str,
        exp_time: str | None = None,
    ) -> dict:
        """Получение сводных рыночных данных по опционам.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-option-market-data
        """
        params = {
            "instFamily": inst_family,
            "expTime": exp_time,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/opt-summary",
            params=params,
        )

    async def get_discount_rate_quota(
        self,
        ccy: str | None = None,
        discount_lv: str | None = None,
    ) -> dict:
        """Получение ставок скидок и беспроцентных квот.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-discount-rate-and-interest-free-quota
        """
        params = {
            "ccy": ccy,
            "discountLv": discount_lv,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/discount-rate-interest-free-quota",
            params=params,
        )

    async def get_system_time(self) -> dict:
        """Получение времени сервера OKX.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-system-time
        """
        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/time",
        )

    async def get_mark_price(
        self,
        inst_type: Literal["MARGIN", "SWAP", "FUTURES", "OPTION"],
        inst_family: str | None = None,
        inst_id: str | None = None,
    ) -> dict:
        """Получение маржинальной цены инструмента.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-mark-price
        """
        params = {
            "instType": inst_type,
            "instFamily": inst_family,
            "instId": inst_id,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/mark-price",
            params=params,
        )

    async def get_position_tiers(
        self,
        inst_type: Literal["MARGIN", "SWAP", "FUTURES", "OPTION"],
        td_mode: Literal["cross", "isolated"],
        inst_family: str | None = None,
        inst_id: str | None = None,
        ccy: str | None = None,
        tier: str | None = None,
    ) -> dict:
        """Получение уровней позиций и допустимого плеча.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-position-tiers
        """
        params = {
            "instType": inst_type,
            "tdMode": td_mode,
            "instFamily": inst_family,
            "instId": inst_id,
            "ccy": ccy,
            "tier": tier,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/position-tiers",
            params=params,
        )

    async def get_interest_rate_loan_quota(self) -> dict:
        """Получение ставок и квот заимствования.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-interest-rate-and-loan-quota
        """
        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/interest-rate-loan-quota",
        )

    async def get_underlying(self, inst_type: Literal["SWAP", "FUTURES", "OPTION"]) -> dict:
        """Получение списка базовых активов по типу инструмента.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-underlying
        """
        params = {
            "instType": inst_type,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/underlying",
            params=params,
        )

    async def get_insurance_fund(
        self,
        inst_type: Literal["MARGIN", "SWAP", "FUTURES", "OPTION"],
        type_: Literal[
            "regular_update",
            "liquidation_balance_deposit",
            "bankruptcy_loss",
            "platform_revenue",
            "adl",
        ]
        | None = None,
        inst_family: str | None = None,
        ccy: str | None = None,
        before: int | None = None,
        after: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение данных страхового фонда.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-insurance-fund
        """
        params = {
            "instType": inst_type,
            "type": type_,
            "instFamily": inst_family,
            "ccy": ccy,
            "before": before,
            "after": after,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/insurance-fund",
            params=params,
        )

    async def convert_contract_coin(
        self,
        inst_id: str,
        sz: NumberLike,
        type_: Literal["1", "2"] | None = None,
        px: NumberLike | None = None,
        unit: Literal["coin", "usds"] | None = None,
        op_type: Literal["open", "close"] | None = None,
    ) -> dict:
        """Конвертация размера контракта и количества монет.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-unit-convert
        """
        params = {
            "type": type_,
            "instId": inst_id,
            "sz": sz,
            "px": px,
            "unit": unit,
            "opType": op_type,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/convert-contract-coin",
            params=params,
        )

    async def get_option_tick_bands(self, inst_family: str | None = None) -> dict:
        """Получение доступных ценовых диапазонов опционов.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-option-tick-bands
        """
        params = {
            "instType": "OPTION",
            "instFamily": inst_family,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/instrument-tick-bands",
            params=params,
        )

    async def get_premium_history(
        self,
        inst_id: str,
        after: int | None = None,
        before: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение истории премии индекса за полгода.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-premium-history
        """
        params = {
            "instId": inst_id,
            "after": after,
            "before": before,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/premium-history",
            params=params,
        )

    async def get_index_tickers(
        self,
        quote_ccy: str | None = None,
        inst_id: str | None = None,
    ) -> dict:
        """Получение индексов и ключевых метрик по ним.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-index-tickers
        """
        params = {
            "quoteCcy": quote_ccy,
            "instId": inst_id,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/market/index-tickers",
            params=params,
        )

    async def get_index_candlesticks(
        self,
        inst_id: str,
        after: int | None = None,
        before: int | None = None,
        bar: str | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение свечей по индексным значениям.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-index-candlesticks
        """
        params = {
            "instId": inst_id,
            "after": after,
            "before": before,
            "bar": bar,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/market/index-candles",
            params=params,
        )

    async def get_index_candlesticks_history(
        self,
        inst_id: str,
        after: int | None = None,
        before: int | None = None,
        bar: str | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение исторических свечей по индексам.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-index-candlesticks-history
        """
        params = {
            "instId": inst_id,
            "after": after,
            "before": before,
            "bar": bar,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/market/history-index-candles",
            params=params,
        )

    async def get_mark_price_candlesticks(
        self,
        inst_id: str,
        after: int | None = None,
        before: int | None = None,
        bar: str | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение свечей по маржинальной цене инструмента.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-mark-price-candlesticks
        """
        params = {
            "instId": inst_id,
            "after": after,
            "before": before,
            "bar": bar,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/market/mark-price-candles",
            params=params,
        )

    async def get_mark_price_candlesticks_history(
        self,
        inst_id: str,
        after: int | None = None,
        before: int | None = None,
        bar: str | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение исторических свечей по маржинальной цене.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-mark-price-candlesticks-history
        """
        params = {
            "instId": inst_id,
            "after": after,
            "before": before,
            "bar": bar,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/market/history-mark-price-candles",
            params=params,
        )

    async def get_exchange_rate(self) -> dict:
        """Получение средневзвешенного курса USD/CNY за две недели.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-exchange-rate
        """
        return await self._make_request(
            "GET",
            endpoint="/api/v5/market/exchange-rate",
        )

    async def get_index_components(self, index: str) -> dict:
        """Получение состава выбранного индекса.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-index-components
        """
        params = {
            "index": index,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/market/index-components",
            params=params,
        )

    async def get_economic_calendar(
        self,
        region: str | None = None,
        importance: Literal["1", "2", "3"] | None = None,
        before: int | None = None,
        after: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение данных макроэкономического календаря.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-economic-calendar-data
        """
        params = {
            "region": region,
            "importance": importance,
            "before": before,
            "after": after,
            "limit": limit,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/economic-calendar",
            params=params,
            signed=True,
        )

    async def get_market_data_history(
        self,
        module: Literal["1", "2", "3", "6"],
        inst_type: Literal["SPOT", "FUTURES", "SWAP", "OPTION"],
        begin: int,
        end: int,
        inst_id_list: str | None = None,
        inst_family_list: str | None = None,
        date_aggr_type: Literal["daily", "monthly"] = "daily",
    ) -> dict:
        """Получение ссылок на исторические рыночные данные OKX.

        https://www.okx.com/docs-v5/en/#public-data-rest-api-get-historical-market-data
        """
        params = {
            "module": module,
            "instType": inst_type,
            "instIdList": inst_id_list,
            "instFamilyList": inst_family_list,
            "dateAggrType": date_aggr_type,
            "begin": begin,
            "end": end,
        }

        return await self._make_request(
            "GET",
            endpoint="/api/v5/public/market-data-history",
            params=params,
        )



__all__ = ["Client"]

import time
from typing import Any, Literal, Self

import aiohttp
import msgpack
from eth_account import Account
from eth_account.messages import encode_typed_data
from eth_account.signers.local import LocalAccount
from eth_utils.conversions import to_hex
from eth_utils.crypto import keccak

from unicex._base import BaseClient
from unicex.exceptions import NotAuthorized
from unicex.types import LoggerLike, NumberLike
from unicex.utils import filter_params

# Authentication


def _l1_payload(phantom_agent: dict[str, Any]) -> dict[str, Any]:
    """Формирует EIP-712 payload для подписи "агента".

    Простыми словами:
    Это упаковка данных в формат, который кошелёк сможет подписать.
    В Ethereum есть стандарт EIP-712 — "structured data signing".
    Он позволяет подписывать не просто строку, а структуру (объект),
    чтобы потом её можно было проверить.

    Пример:
        >>> phantom = {"source": "a", "connectionId": b"1234...."}
        >>> _l1_payload(phantom)
        {...сложный словарь...}

    Параметры:
        phantom_agent (dict): объект с полями:
            - source (str): откуда пришёл агент ("a" для mainnet, "b" для testnet)
            - connectionId (bytes32): уникальный ID (обычно хэш)

    Возвращает:
        dict: структура для подписи через EIP-712
    """
    return {
        "domain": {
            "chainId": 1337,
            "name": "Exchange",
            "verifyingContract": "0x0000000000000000000000000000000000000000",
            "version": "1",
        },
        "types": {
            "Agent": [
                {"name": "source", "type": "string"},
                {"name": "connectionId", "type": "bytes32"},
            ],
            "EIP712Domain": [
                {"name": "name", "type": "string"},
                {"name": "version", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "verifyingContract", "type": "address"},
            ],
        },
        "primaryType": "Agent",
        "message": phantom_agent,
    }


def _address_to_bytes(address: str) -> bytes:
    r"""Переводит Ethereum-адрес в байты.

    Простыми словами:
    Берём строку вида "0xABC123..." и превращаем её в бинарные данные.
    Это нужно, потому что внутри подписи адрес должен храниться как массив байтов.

    Пример:
        >>> _address_to_bytes("0x0000000000000000000000000000000000000001")
        b'\\x00...\\x01'

    Параметры:
        address (str): строковый Ethereum-адрес, с "0x" или без.

    Возвращает:
        bytes: бинарное представление адреса.
    """
    return bytes.fromhex(address[2:] if address.startswith("0x") else address)


def _construct_phantom_agent(hash: bytes, is_mainnet: bool) -> dict[str, Any]:
    r"""Собирает объект "phantom_agent".

    Простыми словами:
    Это кусочек данных, который будет подписываться.
    В нём указывается:
    - источник ("a" если это mainnet, "b" если не mainnet)
    - connectionId — хэш действий.

    Пример:
        >>> _construct_phantom_agent(b"\\x01" * 32, True)
        {"source": "a", "connectionId": b"\\x01"*32}

    Параметры:
        hash (bytes): хэш действия (32 байта).
        is_mainnet (bool): True если сеть основная, False если тестовая.

    Возвращает:
        dict: объект phantom_agent.
    """
    return {"source": "a" if is_mainnet else "b", "connectionId": hash}


def _action_hash(
    action: dict[str, Any],
    vault_address: str | None,
    nonce: int,
    expires_after: int | None,
) -> bytes:
    r"""Строит хэш действия.

    Простыми словами:
    Берём действие (например, ордер), сериализуем его через msgpack,
    добавляем nonce, адрес хранилища (если есть), срок действия,
    и всё это хэшируем keccak256.
    Получается уникальный "отпечаток" действия.

    Пример:
        >>> action = {"type": "order", "amount": 1}
        >>> _action_hash(action, None, 42, None)
        b"\\xab...\\xff"   # 32 байта

    Параметры:
        action (dict): описание действия (например, ордер).
        vault_address (str | None): адрес кошелька/контракта, если есть.
        nonce (int): уникальный счётчик (чтобы нельзя было повторить действие).
        expires_after (int | None): время (в секундах), когда действие протухает.

    Возвращает:
        bytes: 32-байтовый хэш (keccak256).
    """
    data = msgpack.packb(action)
    data += nonce.to_bytes(8, "big")  # type: ignore
    if vault_address is None:
        data += b"\x00"
    else:
        data += b"\x01"
        data += _address_to_bytes(vault_address)
    if expires_after is not None:
        data += b"\x00"
        data += expires_after.to_bytes(8, "big")
    return keccak(data)


def _sign_inner(wallet: LocalAccount, data: dict[str, Any]) -> dict[str, Any]:
    """Подписывает данные EIP-712 через кошелёк.

    Простыми словами:
    Берём структуру (payload), кодируем её в формат EIP-712
    и просим кошелёк подписать.
    Возвращаем r, s, v — стандартные параметры Ethereum-подписи.

    Пример:
        >>> _sign_inner(wallet, {...})
        {"r": "0x...", "s": "0x...", "v": 27}

    Параметры:
        wallet (LocalAccount): объект кошелька.
        data (dict): структура для подписи (EIP-712).

    Возвращает:
        dict:
            - r (str): часть подписи
            - s (str): часть подписи
            - v (int): "восстановитель" (27 или 28 обычно)
    """
    structured_data = encode_typed_data(full_message=data)
    signed = wallet.sign_message(structured_data)
    return {"r": to_hex(signed["r"]), "s": to_hex(signed["s"]), "v": signed["v"]}


def _sign_l1_action(
    wallet: LocalAccount,
    action: dict[str, Any],
    active_pool: str | None,
    nonce: int,
    expires_after: int | None,
    is_mainnet: bool = True,
) -> dict[str, Any]:
    """Подписывает действие для L1 (основного уровня).

    Простыми словами:
    Это конечная функция, которая собирает всё:
    - делает хэш действия
    - строит phantom_agent
    - формирует payload для подписи
    - подписывает его через кошелёк
    Возвращает r, s, v.

    Пример:
        >>> _sign_l1_action(wallet, {"type": "order"}, None, 1, None, True)
        {"r": "0x...", "s": "0x...", "v": 27}

    Параметры:
        wallet (LocalAccount): объект кошелька.
        action (dict): действие (например, ордер).
        active_pool (str | None): адрес пула (если нужен).
        nonce (int): уникальный номер действия.
        expires_after (int | None): срок жизни действия.
        is_mainnet (bool): True — основная сеть, False — тестовая.

    Возвращает:
        dict:
            - r (str)
            - s (str)
            - v (int)
    """
    hash = _action_hash(action, active_pool, nonce, expires_after)
    phantom_agent = _construct_phantom_agent(hash, is_mainnet)
    data = _l1_payload(phantom_agent)
    return _sign_inner(wallet, data)


def _user_signed_payload(
    primary_type: str,
    payload_types: list[dict[str, str]],
    action: dict[str, Any],
) -> dict[str, Any]:
    """Формирует EIP-712 payload для "user-signed" подписи.

    Простыми словами:
    Это структура для подписи действий от лица пользователя.
    В отличие от L1-подписи (где используется phantom-агент),
    здесь кошелёк подписывает само действие напрямую.

    ChainID и домен указываются специально, чтобы предотвратить
    повторное воспроизведение (replay) на других цепях.

    Пример:
        >>> payload = _user_signed_payload(
        ...     primary_type="Withdraw",
        ...     payload_types=[{"name": "amount", "type": "uint256"}],
        ...     action={"amount": 100, "signatureChainId": "0x66eee"},
        ... )
        >>> payload.keys()
        dict_keys(["domain", "types", "primaryType", "message"])

    Параметры:
        primary_type (str): Основной тип сообщения, например `"Withdraw"`, `"Transfer"`.
        payload_types (list[dict]): Список полей EIP-712 типа, например:
            `[{"name": "amount", "type": "uint256"}, {"name": "recipient", "type": "address"}]`
        action (dict): Само сообщение (поля, которые будут подписаны).
            Должно содержать `"signatureChainId"` (строку в hex).

    Возвращает:
        dict: Полностью готовая структура EIP-712 для подписи кошельком.
    """
    chain_id = int(action["signatureChainId"], 16)
    return {
        "domain": {
            "name": "HyperliquidSignTransaction",
            "version": "1",
            "chainId": chain_id,
            "verifyingContract": "0x0000000000000000000000000000000000000000",
        },
        "types": {
            primary_type: payload_types,
            "EIP712Domain": [
                {"name": "name", "type": "string"},
                {"name": "version", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "verifyingContract", "type": "address"},
            ],
        },
        "primaryType": primary_type,
        "message": action,
    }


def _sign_user_signed_action(
    wallet: LocalAccount,
    action: dict[str, Any],
    payload_types: list[dict[str, str]],
    primary_type: str,
    is_mainnet: bool,
) -> dict[str, Any]:
    """Подписывает действие "user-signed" формата (EIP-712).

    Простыми словами:
    Это альтернатива L1-подписи. Здесь сам пользователь (его кошелёк)
    подписывает объект `action`, который будет исполнен Hyperliquid.
    Отличие от L1: используется другой `chainId` (0x66eee) и поле `"hyperliquidChain"`,
    чтобы действие нельзя было воспроизвести в другой сети.

    Пример:
        >>> action = {"amount": 100}
        >>> payload_types = [{"name": "amount", "type": "uint256"}]
        >>> sig = _sign_user_signed_action(wallet, action, payload_types, "Withdraw", True)
        >>> sig.keys()
        dict_keys(["r", "s", "v"])

    Параметры:
        wallet (LocalAccount): Объект кошелька, созданный через `Account.from_key(...)`.
        action (dict): Содержимое действия, которое нужно подписать.
        payload_types (list[dict[str, str]]): Описание полей типа для EIP-712.
        primary_type (str): Основное имя типа, например `"Withdraw"`, `"Transfer"`.
        is_mainnet (bool): True — подпись для основной сети, False — для тестовой.

    Возвращает:
        dict:
            - r (str): первая часть подписи
            - s (str): вторая часть подписи
            - v (int): "восстановитель" подписи (27 или 28)
    """
    # signatureChainId — цепочка, через которую кошелёк делает подпись (не Hyperliquid chain)
    action["signatureChainId"] = "0x66eee"
    # hyperliquidChain — фактическая среда исполнения
    action["hyperliquidChain"] = "Mainnet" if is_mainnet else "Testnet"

    data = _user_signed_payload(primary_type, payload_types, action)
    return _sign_inner(wallet, data)


USD_SEND_SIGN_TYPES = [
    {"name": "hyperliquidChain", "type": "string"},
    {"name": "destination", "type": "string"},
    {"name": "amount", "type": "string"},
    {"name": "time", "type": "uint64"},
]

SPOT_TRANSFER_SIGN_TYPES = [
    {"name": "hyperliquidChain", "type": "string"},
    {"name": "destination", "type": "string"},
    {"name": "token", "type": "string"},
    {"name": "amount", "type": "string"},
    {"name": "time", "type": "uint64"},
]

WITHDRAW_SIGN_TYPES = [
    {"name": "hyperliquidChain", "type": "string"},
    {"name": "destination", "type": "string"},
    {"name": "amount", "type": "string"},
    {"name": "time", "type": "uint64"},
]

USD_CLASS_TRANSFER_SIGN_TYPES = [
    {"name": "hyperliquidChain", "type": "string"},
    {"name": "amount", "type": "string"},
    {"name": "toPerp", "type": "bool"},
    {"name": "nonce", "type": "uint64"},
]

SEND_ASSET_SIGN_TYPES = [
    {"name": "hyperliquidChain", "type": "string"},
    {"name": "destination", "type": "string"},
    {"name": "sourceDex", "type": "string"},
    {"name": "destinationDex", "type": "string"},
    {"name": "token", "type": "string"},
    {"name": "amount", "type": "string"},
    {"name": "fromSubAccount", "type": "string"},
    {"name": "nonce", "type": "uint64"},
]

STAKING_SIGN_TYPES = [
    {"name": "hyperliquidChain", "type": "string"},
    {"name": "wei", "type": "uint64"},
    {"name": "nonce", "type": "uint64"},
]

TOKEN_DELEGATE_TYPES = [
    {"name": "hyperliquidChain", "type": "string"},
    {"name": "validator", "type": "address"},
    {"name": "wei", "type": "uint64"},
    {"name": "isUndelegate", "type": "bool"},
    {"name": "nonce", "type": "uint64"},
]

APPROVE_AGENT_SIGN_TYPES = [
    {"name": "hyperliquidChain", "type": "string"},
    {"name": "agentAddress", "type": "address"},
    {"name": "agentName", "type": "string"},
    {"name": "nonce", "type": "uint64"},
]

APPROVE_BUILDER_FEE_SIGN_TYPES = [
    {"name": "hyperliquidChain", "type": "string"},
    {"name": "maxFeeRate", "type": "string"},
    {"name": "builder", "type": "address"},
    {"name": "nonce", "type": "uint64"},
]


class Client(BaseClient):
    """Клиент для работы с Hyperliquid API."""

    _BASE_URL = "https://api.hyperliquid.xyz"
    """Базовый URL для REST API Hyperliquid."""

    _BASE_HEADERS = {"Content-Type": "application/json"}

    def __init__(
        self,
        session: aiohttp.ClientSession,
        private_key: str | bytes | None = None,
        wallet_address: str | None = None,
        vault_address: str | None = None,
        logger: LoggerLike | None = None,
        max_retries: int = 3,
        retry_delay: int | float = 0.1,
        proxies: list[str] | None = None,
        timeout: int = 10,
    ) -> None:
        """Инициализация клиента.

        Параметры:
            session (`aiohttp.ClientSession`): Сессия для выполнения HTTP‑запросов.
            private_key (`str | bytes | None`): Приватный ключ API для аутентификации (Hyperliquid).
            wallet_address (`str | None`): Адрес кошелька для аутентификации (Hyperliquid).
            vault_address (`str | None`): Адрес хранилища для аутентификации (Hyperliquid).
            logger (`LoggerLike | None`): Логгер для вывода информации.
            max_retries (`int`): Максимальное количество повторных попыток запроса.
            retry_delay (`int | float`): Задержка между повторными попытками, сек.
            proxies (`list[str] | None`): Список HTTP(S)‑прокси для циклического использования.
            timeout (`int`): Максимальное время ожидания ответа от сервера, сек.
        """
        super().__init__(
            session,
            None,
            None,
            None,
            logger,
            max_retries,
            retry_delay,
            proxies,
            timeout,
        )
        self._vault_address = vault_address
        self._wallet_address = wallet_address
        self._wallet: LocalAccount | None = None
        if private_key is not None:
            # private_key может быть в hex-строке ("0x...") или в байтах
            self._wallet = Account.from_key(private_key)

    @classmethod
    async def create(
        cls,
        private_key: str | bytes | None = None,
        wallet_address: str | None = None,
        vault_address: str | None = None,
        session: aiohttp.ClientSession | None = None,
        logger: LoggerLike | None = None,
        max_retries: int = 3,
        retry_delay: int | float = 0.1,
        proxies: list[str] | None = None,
        timeout: int = 10,
    ) -> Self:
        """Создаёт инстанцию клиента.

        Параметры:
            private_key (`str | bytes | None`): Приватный ключ для подписи запросов.
            wallet_address (`str | None`): Адрес кошелька для подписи запросов.
            vault_address (`str | None`): Адрес валита для подписи запросов.
            session (`aiohttp.ClientSession | None`): Сессия для HTTP‑запросов (если не передана, будет создана).
            logger (`LoggerLike | None`): Логгер для вывода информации.
            max_retries (`int`): Максимум повторов при ошибках запроса.
            retry_delay (`int | float`): Задержка между повторами, сек.
            proxies (`list[str] | None`): Список HTTP(S)‑прокси.
            timeout (`int`): Таймаут ответа сервера, сек.

        Возвращает:
            `Self`: Созданный экземпляр клиента.
        """
        return cls(
            private_key=private_key,
            wallet_address=wallet_address,
            vault_address=vault_address,
            session=session or aiohttp.ClientSession(),
            logger=logger,
            max_retries=max_retries,
            retry_delay=retry_delay,
            proxies=proxies,
            timeout=timeout,
        )

    async def _make_request(
        self,
        method: Literal["GET", "POST"],
        endpoint: str,
        *,
        data: dict[str, Any] | None = None,
    ) -> Any:
        """Создаёт HTTP-запрос к Hyperliquid API.

        Параметры:
            method (`Literal["GET", "POST"]`): HTTP-метод запроса.
            endpoint (`str`): Относительный путь эндпоинта.
            data (`dict[str, Any] | None`): Тело запроса для POST-запросов.

        Возвращает:
          `Any`: Ответ Hyperliquid API.
        """
        filtered_data = filter_params(data) if data is not None else None

        return await super()._make_request(
            method=method,
            url=self._BASE_URL + endpoint,
            headers=self._BASE_HEADERS,
            data=filtered_data,
        )

    async def _post_request(self, endpoint: str, data: dict) -> Any:
        """Создание POST-запроса к Hyperliquid API."""
        return await self._make_request("POST", endpoint, data=data)

    async def request(
        self,
        endpoint: str,
        *,
        data: dict,
    ) -> Any:
        """Специальный метод для выполнения запросов на эндпоинты, которые не обернуты в клиенте.

        Параметры:
            endpoint (`str`): Относительный путь эндпоинта (например, `/info`).
            data (`dict[str, Any]`): JSON-тело запроса для POST-запросов.

        Возвращает:
          `Any`: Ответ Hyperliquid API.
        """
        return await self._post_request(endpoint, data=data)

    # topic: Info endpoint
    # topic: Futures

    async def perp_dexs(self) -> list[dict | None]:
        """Получение списка доступных перпетуальных DEX.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint/perpetuals#retrieve-perpetuals-metadata-universe-and-margin-tables
        """
        payload = {"type": "perpDexs"}

        return await self._post_request("/info", data=payload)

    async def perp_metadata(self, dex: str | None = None) -> dict:
        """Получение метаданных по перпетуальным контрактам.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint/perpetuals#retrieve-perpetuals-metadata-universe-and-margin-tables
        """
        payload = {
            "type": "meta",
            "dex": dex,
        }

        return await self._post_request("/info", data=payload)

    async def perp_meta_and_asset_contexts(self) -> list[Any]:
        """Получение метаданных и контекстов активов перпетуальных контрактов.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint/perpetuals#retrieve-perpetuals-asset-contexts-includes-mark-price-current-funding-open-interest-etc
        """
        payload = {"type": "metaAndAssetCtxs"}

        return await self._post_request("/info", data=payload)

    async def perp_account_summary(self, user: str, dex: str | None = None) -> dict:
        """Получение сводной информации по аккаунту пользователя в перпетуалах.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint/perpetuals#retrieve-users-perpetuals-account-summary
        """
        payload = {
            "type": "clearinghouseState",
            "user": user,
            "dex": dex,
        }

        return await self._post_request("/info", data=payload)

    async def perp_user_funding_history(
        self,
        user: str,
        start_time: int,
        end_time: int | None = None,
    ) -> list[dict]:
        """Получение истории фондирования пользователя.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint/perpetuals#retrieve-a-users-funding-history-or-non-funding-ledger-updates
        """
        payload = {
            "type": "userFunding",
            "user": user,
            "startTime": start_time,
            "endTime": end_time,
        }

        return await self._post_request("/info", data=payload)

    async def perp_user_non_funding_ledger_updates(
        self,
        user: str,
        start_time: int,
        end_time: int | None = None,
    ) -> list[dict]:
        """Получение нефондировочных обновлений леджера пользователя.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint/perpetuals#retrieve-a-users-funding-history-or-non-funding-ledger-updates
        """
        payload = {
            "type": "userNonFundingLedgerUpdates",
            "user": user,
            "startTime": start_time,
            "endTime": end_time,
        }

        return await self._post_request("/info", data=payload)

    async def perp_funding_history(
        self,
        coin: str,
        start_time: int,
        end_time: int | None = None,
    ) -> list[dict]:
        """Получение исторических ставок фондирования по монете.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint/perpetuals#retrieve-historical-funding-rates
        """
        payload = {
            "type": "fundingHistory",
            "coin": coin,
            "startTime": start_time,
            "endTime": end_time,
        }

        return await self._post_request("/info", data=payload)

    async def perp_predicted_fundings(self) -> list[list[Any]]:
        """Получение предсказанных ставок фондирования по площадкам.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint/perpetuals#retrieve-predicted-funding-rates-for-different-venues
        """
        payload = {"type": "predictedFundings"}

        return await self._post_request("/info", data=payload)

    async def perps_at_open_interest_cap(self) -> list[str]:
        """Получение списка перпетуалов с достигнутым лимитом открытого интереса.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint/perpetuals#query-perps-at-open-interest-caps
        """
        payload = {"type": "perpsAtOpenInterestCap"}

        return await self._post_request("/info", data=payload)

    async def perp_deploy_auction_status(self) -> dict:
        """Получение статуса аукциона развёртывания перпетуалов.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint/perpetuals#retrieve-information-about-the-perp-deploy-auction
        """
        payload = {"type": "perpDeployAuctionStatus"}

        return await self._post_request("/info", data=payload)

    async def perp_active_asset_data(self, user: str, coin: str) -> dict:
        """Получение активных параметров актива пользователя в перпетуалах.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint/perpetuals#retrieve-users-active-asset-data
        """
        payload = {
            "type": "activeAssetData",
            "user": user,
            "coin": coin,
        }

        return await self._post_request("/info", data=payload)

    async def perp_dex_limits(self, dex: str) -> dict:
        """Получение лимитов для перпетуального DEX, развёрнутого билдерами.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint/perpetuals#retrieve-builder-deployed-perp-market-limits
        """
        payload = {
            "type": "perpDexLimits",
            "dex": dex,
        }

        return await self._post_request("/info", data=payload)

    async def open_orders(self, user: str, dex: str | None = None) -> list[dict]:
        """Получение списка открытых ордеров пользователя.

        https://docs.chainstack.com/reference/hyperliquid-info-openorders
        """
        payload = {
            "type": "openOrders",
            "user": user,
            "dex": dex,
        }

        return await self._post_request("/info", data=payload)

    async def exchange_status(self) -> dict:
        """Получение текущего статуса биржи Hyperliquid.

        https://docs.chainstack.com/reference/hyperliquid-info-exchangestatus
        """
        payload = {"type": "exchangeStatus"}

        return await self._post_request("/info", data=payload)

    async def frontend_open_orders(self, user: str, dex: str | None = None) -> list[dict]:
        """Получение открытых ордеров в формате фронтенда.

        https://docs.chainstack.com/reference/hyperliquid-info-frontendopenorders
        """
        payload = {
            "type": "frontendOpenOrders",
            "user": user,
            "dex": dex,
        }

        return await self._post_request("/info", data=payload)

    async def liquidatable(self, user: str) -> dict:
        """Проверка, подлежит ли аккаунт пользователя ликвидации.

        https://docs.chainstack.com/reference/hyperliquid-info-liquidatable
        """
        payload = {
            "type": "liquidatable",
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def max_market_order_ntls(self) -> list[dict]:
        """Получение максимальных объёмов рыночных ордеров по активам.

        https://docs.chainstack.com/reference/hyperliquid-info-maxmarketorderntls
        """
        payload = {"type": "maxMarketOrderNtls"}

        return await self._post_request("/info", data=payload)

    async def vault_summaries(self) -> list[dict]:
        """Получение сводки по всем доступным вултам (vaults).

        https://docs.chainstack.com/reference/hyperliquid-info-vaultsummaries
        """
        payload = {"type": "vaultSummaries"}

        return await self._post_request("/info", data=payload)

    async def user_vault_equities(self, user: str) -> list[dict]:
        """Получение данных об инвестициях пользователя в вулты (vaults).

        https://docs.chainstack.com/reference/hyperliquid-info-uservaultequities
        """
        payload = {
            "type": "userVaultEquities",
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def leading_vaults(self, user: str) -> list[dict]:
        """Получение списка вултов (vaults), которыми управляет пользователь.

        https://docs.chainstack.com/reference/hyperliquid-info-leadingvaults
        """
        payload = {
            "type": "leadingVaults",
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def extra_agents(self, user: str) -> list[dict]:
        """Получение списка дополнительных агентов пользователя.

        https://docs.chainstack.com/reference/hyperliquid-info-extraagents
        """
        payload = {
            "type": "extraAgents",
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def sub_accounts(self, user: str) -> list[dict]:
        """Получение списка саб-аккаунтов пользователя.

        https://docs.chainstack.com/reference/hyperliquid-info-subaccounts
        """
        payload = {
            "type": "subAccounts",
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def user_fees(self, user: str) -> dict:
        """Получение информации о торговых комиссиях пользователя.

        https://docs.chainstack.com/reference/hyperliquid-info-userfees
        """
        payload = {
            "type": "userFees",
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def user_rate_limit(self, user: str) -> dict:
        """Получение сведений о лимитах запросов пользователя.

        https://docs.chainstack.com/reference/hyperliquid-info-userratelimit
        """
        payload = {
            "type": "userRateLimit",
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def delegations(self, user: str) -> list[dict]:
        """Получение списка делегаций пользователя.

        https://docs.chainstack.com/reference/hyperliquid-info-delegations
        """
        payload = {
            "type": "delegations",
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def delegator_summary(self, user: str) -> dict:
        """Получение сводки по делегациям пользователя.

        https://docs.chainstack.com/reference/hyperliquid-info-delegator-summary
        """
        payload = {
            "type": "delegatorSummary",
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def max_builder_fee(self, user: str, builder: str) -> int:
        """Получение максимальной комиссии билдера, одобренной пользователем.

        https://docs.chainstack.com/reference/hyperliquid-info-max-builder-fee
        """
        payload = {
            "type": "maxBuilderFee",
            "user": user,
            "builder": builder,
        }

        return await self._post_request("/info", data=payload)

    async def user_to_multi_sig_signers(self, user: str) -> list[str]:
        """Получение списка подписантов мультисиг-кошелька пользователя.

        https://docs.chainstack.com/reference/hyperliquid-info-user-to-multi-sig-signers
        """
        payload = {
            "type": "userToMultiSigSigners",
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def user_role(self, user: str) -> dict:
        """Получение информации о роли пользователя в системе.

        https://docs.chainstack.com/reference/hyperliquid-info-user-role
        """
        payload = {
            "type": "userRole",
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def validator_l1_votes(self) -> list[dict]:
        """Получение сведений о голосах валидаторов на L1.

        https://docs.chainstack.com/reference/hyperliquid-info-validator-l1-votes
        """
        payload = {"type": "validatorL1Votes"}

        return await self._post_request("/info", data=payload)

    async def web_data2(self) -> dict:
        """Получение агрегированных данных для веб-интерфейса.

        https://docs.chainstack.com/reference/hyperliquid-info-web-data2
        """
        payload = {"type": "webData2"}

        return await self._post_request("/info", data=payload)

    async def all_mids(self, dex: str | None = None) -> dict:
        """Получение текущих средних цен по всем активам.

        https://docs.chainstack.com/reference/hyperliquid-info-allmids
        """
        payload = {
            "type": "allMids",
            "dex": dex,
        }

        return await self._post_request("/info", data=payload)

    async def user_fills(
        self,
        user: str,
        aggregate_by_time: bool | None = None,
    ) -> list[dict]:
        """Получение последних исполнений ордеров пользователя.

        https://docs.chainstack.com/reference/hyperliquid-info-user-fills
        """
        payload = {
            "type": "userFills",
            "user": user,
            "aggregateByTime": aggregate_by_time,
        }

        return await self._post_request("/info", data=payload)

    async def user_fills_by_time(
        self,
        user: str,
        start_time: int,
        end_time: int | None = None,
        aggregate_by_time: bool | None = None,
    ) -> list[dict]:
        """Получение исполнений ордеров пользователя за период.

        https://docs.chainstack.com/reference/hyperliquid-info-user-fills-by-time
        """
        payload = {
            "type": "userFillsByTime",
            "user": user,
            "startTime": start_time,
            "endTime": end_time,
            "aggregateByTime": aggregate_by_time,
        }

        return await self._post_request("/info", data=payload)

    async def order_status(self, user: str, oid: int | str) -> dict:
        """Получение статуса ордера по идентификатору.

        https://docs.chainstack.com/reference/hyperliquid-info-order-status
        """
        payload = {
            "type": "orderStatus",
            "user": user,
            "oid": oid,
        }

        return await self._post_request("/info", data=payload)

    async def l2_book(
        self,
        coin: str,
        n_sig_figs: Literal[2, 3, 4, 5] | None = None,
        mantissa: Literal[1, 2, 5] | None = None,
    ) -> list[list[dict]]:
        """Получение снапшота стакана уровня L2 для актива.

        https://docs.chainstack.com/reference/hyperliquid-info-l2-book
        """
        payload = {
            "type": "l2Book",
            "coin": coin,
            "nSigFigs": n_sig_figs,
            "mantissa": mantissa,
        }

        return await self._post_request("/info", data=payload)

    async def batch_clearinghouse_states(
        self,
        users: list[str],
        dex: str | None = None,
    ) -> list[dict | None]:
        """Получение сводок фьючерсных аккаунтов группы пользователей.

        https://docs.chainstack.com/reference/hyperliquid-info-batch-clearinghouse-states
        """
        payload = {
            "type": "batchClearinghouseStates",
            "users": users,
            "dex": dex,
        }

        return await self._post_request("/info", data=payload)

    async def candle_snapshot(
        self,
        coin: str,
        interval: str,
        start_time: int,
        end_time: int,
    ) -> list[dict]:
        """Получение датасета свечей за указанный период.

        https://docs.chainstack.com/reference/hyperliquid-info-candle-snapshot
        """
        payload = {
            "type": "candleSnapshot",
            "req": {
                "coin": coin,
                "interval": interval,
                "startTime": start_time,
                "endTime": end_time,
            },
        }

        return await self._post_request("/info", data=payload)

    async def historical_orders(self, user: str) -> list[dict]:
        """Получение истории ордеров пользователя.

        https://docs.chainstack.com/reference/hyperliquid-info-historical-orders
        """
        payload = {
            "type": "historicalOrders",
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def user_twap_slice_fills(self, user: str) -> list[dict]:
        """Получение последних TWAP-исполнений пользователя.

        https://docs.chainstack.com/reference/hyperliquid-info-user-twap-slice-fills
        """
        payload = {
            "type": "userTwapSliceFills",
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def recent_trades(self, coin: str) -> list[dict]:
        """Получение последних публичных сделок по активу.

        https://docs.chainstack.com/reference/hyperliquid-info-recent-trades
        """
        payload = {
            "type": "recentTrades",
            "coin": coin,
        }

        return await self._post_request("/info", data=payload)

    async def vault_details(self, vault_address: str, user: str | None = None) -> dict:
        """Получение подробной информации о выбранном вулте.

        https://docs.chainstack.com/reference/hyperliquid-info-vault-details
        """
        payload = {
            "type": "vaultDetails",
            "vaultAddress": vault_address,
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def portfolio(self, user: str) -> list[list[Any]]:
        """Получение данных о производительности портфеля пользователя.

        https://docs.chainstack.com/reference/hyperliquid-info-portfolio
        """
        payload = {
            "type": "portfolio",
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def referral(self, user: str) -> dict:
        """Получение реферальной информации пользователя.

        https://docs.chainstack.com/reference/hyperliquid-info-referral
        """
        payload = {
            "type": "referral",
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def delegator_rewards(self, user: str) -> list[dict]:
        """Получение истории стейкинг-наград пользователя.

        https://docs.chainstack.com/reference/hyperliquid-info-delegator-rewards
        """
        payload = {
            "type": "delegatorRewards",
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def gossip_root_ips(self) -> list[str]:
        """Получение списка узлов для P2P-госсипа.

        https://docs.chainstack.com/reference/hyperliquid-info-gossip-root-ips
        """
        payload = {"type": "gossipRootIps"}

        return await self._post_request("/info", data=payload)

    # topic: Spot

    async def spot_metadata(self) -> dict:
        """Получение спотовой метаинформации о бирже.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint/spot#retrieve-spot-metadata
        """
        payload = {"type": "spotMeta"}

        return await self._post_request("/info", data=payload)

    async def spot_meta_and_asset_contexts(self) -> list[Any]:
        """Получение метаданных и контекстов спотовых активов.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint/spot#retrieve-spot-asset-contexts
        """
        payload = {"type": "spotMetaAndAssetCtxs"}

        return await self._post_request("/info", data=payload)

    async def spot_token_balances(self, user: str) -> dict:
        """Получение балансов токенов пользователя на спотовом рынке.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint/spot#retrieve-a-users-token-balances
        """
        payload = {
            "type": "spotClearinghouseState",
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def spot_deploy_state(self, user: str) -> dict:
        """Получение информации об аукционе развёртывания спотового токена.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint/spot#retrieve-information-about-the-spot-deploy-auction
        """
        payload = {
            "type": "spotDeployState",
            "user": user,
        }

        return await self._post_request("/info", data=payload)

    async def spot_pair_deploy_auction_status(self) -> dict:
        """Получение статуса аукциона развёртывания спотовых пар.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint/spot#retrieve-information-about-the-spot-pair-deploy-auction
        """
        payload = {"type": "spotPairDeployAuctionStatus"}

        return await self._post_request("/info", data=payload)

    async def spot_token_details(self, token_id: str) -> dict:
        """Получение подробной информации о спотовом токене.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint/spot#retrieve-information-about-a-token
        """
        payload = {
            "type": "tokenDetails",
            "tokenId": token_id,
        }

        return await self._post_request("/info", data=payload)

    # topic: Exchange endpoint

    async def place_order(
        self,
        asset: int,
        is_buy: bool,
        size: NumberLike,
        reduce_only: bool,
        order_type: Literal["limit", "trigger"],
        order_body: dict,
        price: NumberLike | None = None,
        client_order_id: str | None = None,
        grouping: Literal["na", "normalTpsl", "positionTpsl"] = "na",
        builder_address: str | None = None,
        builder_fee: int | None = None,
        nonce: int | None = None,
        expires_after: int | None = None,
        vault_address: str | None = None,
    ) -> dict:
        """Выставление ордера на бирже.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#place-an-order
        """
        order_payload = {
            "a": asset,
            "b": is_buy,
            "p": price,
            "s": size,
            "r": reduce_only,
            "t": {order_type: order_body},
        }
        if client_order_id is not None:
            order_payload["c"] = client_order_id

        return await self.batch_place_orders(
            [order_payload],
            grouping=grouping,
            builder_address=builder_address,
            builder_fee=builder_fee,
            nonce=nonce,
            expires_after=expires_after,
            vault_address=vault_address,
        )

    async def batch_place_orders(
        self,
        orders: list[dict[str, Any]],
        grouping: Literal["na", "normalTpsl", "positionTpsl"] = "na",
        builder_address: str | None = None,
        builder_fee: int | None = None,
        nonce: int | None = None,
        expires_after: int | None = None,
        vault_address: str | None = None,
    ) -> dict:
        """Пакетное выставление ордеров на бирже.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#place-an-order
        """
        if not orders:
            raise ValueError("orders must not be empty")
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")
        if builder_address is not None and builder_fee is None:
            raise TypeError("builder_fee is required when builder_address is provided")
        if builder_address is None and builder_fee is not None:
            raise TypeError("builder_address is required when builder_fee is provided")

        required_keys = {"a", "b", "p", "s", "r", "t"}
        normalized_orders = []
        for order in orders:
            missing_keys = required_keys - order.keys()
            if missing_keys:
                missing = ", ".join(sorted(missing_keys))
                raise ValueError(f"order is missing required fields: {missing}")
            normalized = dict(order)
            normalized["p"] = str(normalized["p"])
            normalized["s"] = str(normalized["s"])
            if normalized.get("c") is None:
                normalized.pop("c", None)
            normalized_orders.append(normalized)

        action = {
            "type": "order",
            "orders": normalized_orders,
            "grouping": grouping,
        }
        if builder_address is not None:
            action["builder"] = {"b": builder_address, "f": builder_fee}

        effective_vault = vault_address or self._vault_address
        action_nonce = nonce if nonce is not None else int(time.time() * 1000)
        signature = _sign_l1_action(
            self._wallet,
            action,
            effective_vault,
            action_nonce,
            expires_after,
        )

        payload = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }
        if effective_vault is not None:
            payload["vaultAddress"] = effective_vault
        if expires_after is not None:
            payload["expiresAfter"] = expires_after

        return await self._post_request("/exchange", data=payload)

    async def cancel_order(
        self,
        asset: int,
        order_id: int,
        nonce: int | None = None,
        expires_after: int | None = None,
        vault_address: str | None = None,
    ) -> dict:
        """Отмена ордера по идентификатору.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#cancel-order-s
        """
        return await self.cancel_orders(
            cancels=[{"a": asset, "o": order_id}],
            nonce=nonce,
            expires_after=expires_after,
            vault_address=vault_address,
        )

    async def cancel_orders(
        self,
        cancels: list[dict[str, int | str]],
        nonce: int | None = None,
        expires_after: int | None = None,
        vault_address: str | None = None,
    ) -> dict:
        """Отмена ордеров по идентификатору ордера.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#cancel-order-s
        """
        if not cancels:
            raise ValueError("cancels must not be empty")
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        normalized: list[dict[str, int]] = []
        for cancel in cancels:
            missing_keys = {"a", "o"} - cancel.keys()
            if missing_keys:
                missing = ", ".join(sorted(missing_keys))
                raise ValueError(f"cancel entry is missing required fields: {missing}")
            normalized.append(
                {
                    "a": int(cancel["a"]),
                    "o": int(cancel["o"]),
                }
            )

        action = {"type": "cancel", "cancels": normalized}

        effective_vault = vault_address or self._vault_address
        action_nonce = nonce if nonce is not None else int(time.time() * 1000)
        signature = _sign_l1_action(
            self._wallet,
            action,
            effective_vault,
            action_nonce,
            expires_after,
        )

        payload: dict[str, Any] = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }
        if effective_vault is not None:
            payload["vaultAddress"] = effective_vault
        if expires_after is not None:
            payload["expiresAfter"] = expires_after

        return await self._post_request("/exchange", data=payload)

    async def cancel_order_by_cloid(
        self,
        asset: int,
        client_order_id: str,
        nonce: int | None = None,
        expires_after: int | None = None,
        vault_address: str | None = None,
    ) -> dict:
        """Отмена ордера по клиентскому идентификатору.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#cancel-order-s-by-cloid
        """
        return await self.cancel_orders_by_cloid(
            cancels=[{"asset": asset, "cloid": client_order_id}],
            nonce=nonce,
            expires_after=expires_after,
            vault_address=vault_address,
        )

    async def cancel_orders_by_cloid(
        self,
        cancels: list[dict[str, Any]],
        nonce: int | None = None,
        expires_after: int | None = None,
        vault_address: str | None = None,
    ) -> dict:
        """Отмена ордеров по клиентскому идентификатору.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#cancel-order-s-by-cloid
        """
        if not cancels:
            raise ValueError("cancels must not be empty")
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        normalized: list[dict[str, Any]] = []
        for cancel in cancels:
            missing_keys = {"asset", "cloid"} - cancel.keys()
            if missing_keys:
                missing = ", ".join(sorted(missing_keys))
                raise ValueError(f"cancel entry is missing required fields: {missing}")
            normalized.append(
                {
                    "asset": int(cancel["asset"]),
                    "cloid": str(cancel["cloid"]),
                }
            )

        action = {"type": "cancelByCloid", "cancels": normalized}

        effective_vault = vault_address or self._vault_address
        action_nonce = nonce if nonce is not None else int(time.time() * 1000)
        signature = _sign_l1_action(
            self._wallet,
            action,
            effective_vault,
            action_nonce,
            expires_after,
        )

        payload: dict[str, Any] = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }
        if effective_vault is not None:
            payload["vaultAddress"] = effective_vault
        if expires_after is not None:
            payload["expiresAfter"] = expires_after

        return await self._post_request("/exchange", data=payload)

    async def schedule_cancel(
        self,
        time_ms: int | None = None,
        nonce: int | None = None,
        expires_after: int | None = None,
        vault_address: str | None = None,
    ) -> dict:
        """Планирование массовой отмены ордеров.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#schedule-cancel-dead-mans-switch
        """
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        action: dict[str, Any] = {"type": "scheduleCancel"}
        if time_ms is not None:
            action["time"] = time_ms

        effective_vault = vault_address or self._vault_address
        action_nonce = nonce if nonce is not None else int(time.time() * 1000)
        signature = _sign_l1_action(
            self._wallet,
            action,
            effective_vault,
            action_nonce,
            expires_after,
        )

        payload: dict[str, Any] = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }
        if effective_vault is not None:
            payload["vaultAddress"] = effective_vault
        if expires_after is not None:
            payload["expiresAfter"] = expires_after

        return await self._post_request("/exchange", data=payload)

    async def modify_order(
        self,
        order_id: int | str,
        asset: int,
        is_buy: bool,
        price: NumberLike,
        size: NumberLike,
        reduce_only: bool,
        order_type: Literal["limit", "trigger"],
        order_body: dict[str, Any],
        client_order_id: str | None = None,
        nonce: int | None = None,
        expires_after: int | None = None,
        vault_address: str | None = None,
    ) -> dict:
        """Модификация существующего ордера.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#modify-an-order
        """
        order_payload: dict[str, Any] = {
            "a": asset,
            "b": is_buy,
            "p": str(price),
            "s": str(size),
            "r": reduce_only,
            "t": {order_type: order_body},
        }
        if client_order_id is not None:
            order_payload["c"] = client_order_id

        return await self.batch_modify_orders(
            modifies=[{"oid": order_id, "order": order_payload}],
            nonce=nonce,
            expires_after=expires_after,
            vault_address=vault_address,
        )

    async def batch_modify_orders(
        self,
        modifies: list[dict[str, Any]],
        nonce: int | None = None,
        expires_after: int | None = None,
        vault_address: str | None = None,
    ) -> dict:
        """Пакетная модификация ордеров.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#modify-multiple-orders
        """
        if not modifies:
            raise ValueError("modifies must not be empty")
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        normalized: list[dict[str, Any]] = []
        for modify in modifies:
            missing_keys = {"oid", "order"} - modify.keys()
            if missing_keys:
                missing = ", ".join(sorted(missing_keys))
                raise ValueError(f"modify entry is missing required fields: {missing}")
            order = dict(modify["order"])
            required_order_keys = {"a", "b", "p", "s", "r", "t"}
            missing_order_keys = required_order_keys - order.keys()
            if missing_order_keys:
                missing = ", ".join(sorted(missing_order_keys))
                raise ValueError(f"order payload is missing required fields: {missing}")
            order["p"] = str(order["p"])
            order["s"] = str(order["s"])
            if order.get("c") is None:
                order.pop("c", None)
            normalized.append(
                {
                    "oid": modify["oid"],
                    "order": order,
                }
            )

        action = {"type": "batchModify", "modifies": normalized}

        effective_vault = vault_address or self._vault_address
        action_nonce = nonce if nonce is not None else int(time.time() * 1000)
        signature = _sign_l1_action(
            self._wallet,
            action,
            effective_vault,
            action_nonce,
            expires_after,
        )

        payload: dict[str, Any] = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }
        if effective_vault is not None:
            payload["vaultAddress"] = effective_vault
        if expires_after is not None:
            payload["expiresAfter"] = expires_after

        return await self._post_request("/exchange", data=payload)

    async def update_leverage(
        self,
        asset: int,
        is_cross: bool,
        leverage: int,
        nonce: int | None = None,
        expires_after: int | None = None,
        vault_address: str | None = None,
    ) -> dict:
        """Обновление кредитного плеча по активу.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#update-leverage
        """
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        action = {
            "type": "updateLeverage",
            "asset": asset,
            "isCross": is_cross,
            "leverage": leverage,
        }

        effective_vault = vault_address or self._vault_address
        action_nonce = nonce if nonce is not None else int(time.time() * 1000)
        signature = _sign_l1_action(
            self._wallet,
            action,
            effective_vault,
            action_nonce,
            expires_after,
        )

        payload: dict[str, Any] = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }
        if effective_vault is not None:
            payload["vaultAddress"] = effective_vault
        if expires_after is not None:
            payload["expiresAfter"] = expires_after

        return await self._post_request("/exchange", data=payload)

    async def update_isolated_margin(
        self,
        asset: int,
        is_buy: bool,
        notional_change: int,
        nonce: int | None = None,
        expires_after: int | None = None,
        vault_address: str | None = None,
    ) -> dict:
        """Обновление маржи изолированной позиции.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#update-isolated-margin
        """
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        action = {
            "type": "updateIsolatedMargin",
            "asset": asset,
            "isBuy": is_buy,
            "ntli": notional_change,
        }

        effective_vault = vault_address or self._vault_address
        action_nonce = nonce if nonce is not None else int(time.time() * 1000)
        signature = _sign_l1_action(
            self._wallet,
            action,
            effective_vault,
            action_nonce,
            expires_after,
        )

        payload: dict[str, Any] = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }
        if effective_vault is not None:
            payload["vaultAddress"] = effective_vault
        if expires_after is not None:
            payload["expiresAfter"] = expires_after

        return await self._post_request("/exchange", data=payload)

    async def usd_send(
        self,
        hyperliquid_chain: Literal["Mainnet", "Testnet"],
        signature_chain_id: str,
        destination: str,
        amount: NumberLike,
        time_ms: int,
        nonce: int | None = None,
    ) -> dict:
        """Перевод USDC между пользователями Hyperliquid.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#core-usdc-transfer
        """
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        action = {
            "type": "usdSend",
            "hyperliquidChain": hyperliquid_chain,
            "signatureChainId": signature_chain_id,
            "destination": destination,
            "amount": amount,
            "time": time_ms,
        }
        action_nonce = nonce if nonce is not None else time_ms
        is_mainnet = hyperliquid_chain == "Mainnet"
        signature = _sign_user_signed_action(
            self._wallet,
            action,
            USD_SEND_SIGN_TYPES,
            "HyperliquidTransaction:UsdSend",
            is_mainnet,
        )

        payload = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }

        return await self._post_request("/exchange", data=payload)

    async def spot_send(
        self,
        hyperliquid_chain: Literal["Mainnet", "Testnet"],
        signature_chain_id: str,
        destination: str,
        token: str,
        amount: NumberLike,
        time_ms: int,
        nonce: int | None = None,
    ) -> dict:
        """Перевод спотового актива между пользователями Hyperliquid.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#core-spot-transfer
        """
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        action = {
            "type": "spotSend",
            "hyperliquidChain": hyperliquid_chain,
            "signatureChainId": signature_chain_id,
            "destination": destination,
            "token": token,
            "amount": amount,
            "time": time_ms,
        }
        action_nonce = nonce if nonce is not None else time_ms
        is_mainnet = hyperliquid_chain == "Mainnet"
        signature = _sign_user_signed_action(
            self._wallet,
            action,
            SPOT_TRANSFER_SIGN_TYPES,
            "HyperliquidTransaction:SpotSend",
            is_mainnet,
        )

        payload = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }

        return await self._post_request("/exchange", data=payload)

    async def initiate_withdrawal(
        self,
        hyperliquid_chain: Literal["Mainnet", "Testnet"],
        signature_chain_id: str,
        amount: NumberLike,
        time_ms: int,
        destination: str,
        nonce: int | None = None,
    ) -> dict:
        """Инициация вывода средств на L1.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#initiate-a-withdrawal-request
        """
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        action = {
            "type": "withdraw3",
            "hyperliquidChain": hyperliquid_chain,
            "signatureChainId": signature_chain_id,
            "amount": amount,
            "time": time_ms,
            "destination": destination,
        }
        action_nonce = nonce if nonce is not None else time_ms
        is_mainnet = hyperliquid_chain == "Mainnet"
        signature = _sign_user_signed_action(
            self._wallet,
            action,
            WITHDRAW_SIGN_TYPES,
            "HyperliquidTransaction:Withdraw",
            is_mainnet,
        )

        payload = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }

        return await self._post_request("/exchange", data=payload)

    async def usd_class_transfer(
        self,
        hyperliquid_chain: Literal["Mainnet", "Testnet"],
        signature_chain_id: str,
        amount: NumberLike,
        to_perp: bool,
        subaccount: str | None = None,
    ) -> dict:
        """Перевод USDC между спотовым и перпетуальным аккаунтами.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#transfer-from-spot-account-to-perp-account-and-vice-versa
        """
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        amount_field = amount
        if subaccount is not None:
            amount_field = f"{amount} subaccount:{subaccount}"

        nonce = int(time.time() * 1000)

        action = {
            "type": "usdClassTransfer",
            "hyperliquidChain": hyperliquid_chain,
            "signatureChainId": signature_chain_id,
            "amount": amount_field,
            "toPerp": to_perp,
            "nonce": nonce,
        }
        is_mainnet = hyperliquid_chain == "Mainnet"
        signature = _sign_user_signed_action(
            self._wallet,
            action,
            USD_CLASS_TRANSFER_SIGN_TYPES,
            "HyperliquidTransaction:UsdClassTransfer",
            is_mainnet,
        )

        payload = {
            "action": action,
            "nonce": nonce,
            "signature": signature,
        }

        return await self._post_request("/exchange", data=payload)

    async def send_asset(
        self,
        hyperliquid_chain: Literal["Mainnet", "Testnet"],
        signature_chain_id: str,
        destination: str,
        source_dex: str,
        destination_dex: str,
        token: str,
        amount: NumberLike,
        from_subaccount: str,
        nonce_value: int,
        nonce: int | None = None,
    ) -> dict:
        """Перевод токена между балансами и субаккаунтами (тестнет).

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#send-asset-testnet-only
        """
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        action = {
            "type": "sendAsset",
            "hyperliquidChain": hyperliquid_chain,
            "signatureChainId": signature_chain_id,
            "destination": destination,
            "sourceDex": source_dex,
            "destinationDex": destination_dex,
            "token": token,
            "amount": amount,
            "fromSubAccount": from_subaccount,
            "nonce": nonce_value,
        }
        action_nonce = nonce if nonce is not None else nonce_value
        is_mainnet = hyperliquid_chain == "Mainnet"
        signature = _sign_user_signed_action(
            self._wallet,
            action,
            SEND_ASSET_SIGN_TYPES,
            "HyperliquidTransaction:SendAsset",
            is_mainnet,
        )

        payload = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }

        return await self._post_request("/exchange", data=payload)

    async def staking_deposit(
        self,
        hyperliquid_chain: Literal["Mainnet", "Testnet"],
        signature_chain_id: str,
        wei_amount: int,
        nonce_value: int,
        nonce: int | None = None,
    ) -> dict:
        """Депозит нативного токена в стейкинг.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#deposit-into-staking
        """
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        action = {
            "type": "cDeposit",
            "hyperliquidChain": hyperliquid_chain,
            "signatureChainId": signature_chain_id,
            "wei": wei_amount,
            "nonce": nonce_value,
        }
        action_nonce = nonce if nonce is not None else nonce_value
        is_mainnet = hyperliquid_chain == "Mainnet"
        signature = _sign_user_signed_action(
            self._wallet,
            action,
            STAKING_SIGN_TYPES,
            "HyperliquidTransaction:CDeposit",
            is_mainnet,
        )

        payload = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }

        return await self._post_request("/exchange", data=payload)

    async def staking_withdraw(
        self,
        hyperliquid_chain: Literal["Mainnet", "Testnet"],
        signature_chain_id: str,
        wei_amount: int,
        nonce_value: int,
        nonce: int | None = None,
    ) -> dict:
        """Вывод нативного токена из стейкинга.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#withdraw-from-staking
        """
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        action = {
            "type": "cWithdraw",
            "hyperliquidChain": hyperliquid_chain,
            "signatureChainId": signature_chain_id,
            "wei": wei_amount,
            "nonce": nonce_value,
        }
        action_nonce = nonce if nonce is not None else nonce_value
        is_mainnet = hyperliquid_chain == "Mainnet"
        signature = _sign_user_signed_action(
            self._wallet,
            action,
            STAKING_SIGN_TYPES,
            "HyperliquidTransaction:CWithdraw",
            is_mainnet,
        )

        payload = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }

        return await self._post_request("/exchange", data=payload)

    async def token_delegate(
        self,
        hyperliquid_chain: Literal["Mainnet", "Testnet"],
        signature_chain_id: str,
        validator: str,
        is_undelegate: bool,
        wei_amount: int,
        nonce_value: int,
        nonce: int | None = None,
    ) -> dict:
        """Делегирование или отзыв делегирования нативного токена валидатору.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#delegate-or-undelegate-stake-from-validator
        """
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        action = {
            "type": "tokenDelegate",
            "hyperliquidChain": hyperliquid_chain,
            "signatureChainId": signature_chain_id,
            "validator": validator,
            "isUndelegate": is_undelegate,
            "wei": wei_amount,
            "nonce": nonce_value,
        }
        action_nonce = nonce if nonce is not None else nonce_value
        is_mainnet = hyperliquid_chain == "Mainnet"
        signature = _sign_user_signed_action(
            self._wallet,
            action,
            TOKEN_DELEGATE_TYPES,
            "HyperliquidTransaction:TokenDelegate",
            is_mainnet,
        )

        payload = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }

        return await self._post_request("/exchange", data=payload)

    async def vault_transfer(
        self,
        vault_address: str,
        is_deposit: bool,
        usd: NumberLike,
        nonce: int | None = None,
        expires_after: int | None = None,
        signing_vault_address: str | None = None,
    ) -> dict:
        """Перевод средств в или из хранилища.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#deposit-or-withdraw-from-a-vault
        """
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        action = {
            "type": "vaultTransfer",
            "vaultAddress": vault_address,
            "isDeposit": is_deposit,
            "usd": usd,
        }

        effective_vault = signing_vault_address or self._vault_address
        action_nonce = nonce if nonce is not None else int(time.time() * 1000)
        signature = _sign_l1_action(
            self._wallet,
            action,
            effective_vault,
            action_nonce,
            expires_after,
        )

        payload: dict[str, Any] = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }
        if effective_vault is not None:
            payload["vaultAddress"] = effective_vault
        if expires_after is not None:
            payload["expiresAfter"] = expires_after

        return await self._post_request("/exchange", data=payload)

    async def approve_agent(
        self,
        hyperliquid_chain: Literal["Mainnet", "Testnet"],
        signature_chain_id: str,
        agent_address: str,
        nonce_value: int,
        agent_name: str | None = None,
        nonce: int | None = None,
    ) -> dict:
        """Одобрение API-кошелька.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#approve-an-api-wallet
        """
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        action: dict[str, Any] = {
            "type": "approveAgent",
            "hyperliquidChain": hyperliquid_chain,
            "signatureChainId": signature_chain_id,
            "agentAddress": agent_address,
            "nonce": nonce_value,
        }
        if agent_name is not None:
            action["agentName"] = agent_name
        else:
            action["agentName"] = ""
        action_nonce = nonce if nonce is not None else nonce_value
        is_mainnet = hyperliquid_chain == "Mainnet"
        signature = _sign_user_signed_action(
            self._wallet,
            action,
            APPROVE_AGENT_SIGN_TYPES,
            "HyperliquidTransaction:ApproveAgent",
            is_mainnet,
        )

        payload = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }

        return await self._post_request("/exchange", data=payload)

    async def approve_builder_fee(
        self,
        hyperliquid_chain: Literal["Mainnet", "Testnet"],
        signature_chain_id: str,
        max_fee_rate: str,
        builder_address: str,
        nonce_value: int,
        nonce: int | None = None,
    ) -> dict:
        """Одобрение максимальной комиссии билдера.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#approve-a-builder-fee
        """
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        action = {
            "type": "approveBuilderFee",
            "hyperliquidChain": hyperliquid_chain,
            "signatureChainId": signature_chain_id,
            "maxFeeRate": max_fee_rate,
            "builder": builder_address,
            "nonce": nonce_value,
        }
        action_nonce = nonce if nonce is not None else nonce_value
        is_mainnet = hyperliquid_chain == "Mainnet"
        signature = _sign_user_signed_action(
            self._wallet,
            action,
            APPROVE_BUILDER_FEE_SIGN_TYPES,
            "HyperliquidTransaction:ApproveBuilderFee",
            is_mainnet,
        )

        payload = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }

        return await self._post_request("/exchange", data=payload)

    async def place_twap_order(
        self,
        asset: int,
        is_buy: bool,
        size: NumberLike,
        reduce_only: bool,
        minutes: int,
        randomize: bool,
        nonce: int | None = None,
        expires_after: int | None = None,
        vault_address: str | None = None,
    ) -> dict:
        """Создание TWAP-ордера.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#place-a-twap-order
        """
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        action = {
            "type": "twapOrder",
            "twap": {
                "a": asset,
                "b": is_buy,
                "s": str(size),
                "r": reduce_only,
                "m": minutes,
                "t": randomize,
            },
        }

        effective_vault = vault_address or self._vault_address
        action_nonce = nonce if nonce is not None else int(time.time() * 1000)
        signature = _sign_l1_action(
            self._wallet,
            action,
            effective_vault,
            action_nonce,
            expires_after,
        )

        payload: dict[str, Any] = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }
        if effective_vault is not None:
            payload["vaultAddress"] = effective_vault
        if expires_after is not None:
            payload["expiresAfter"] = expires_after

        return await self._post_request("/exchange", data=payload)

    async def cancel_twap_order(
        self,
        asset: int,
        twap_id: int,
        nonce: int | None = None,
        expires_after: int | None = None,
        vault_address: str | None = None,
    ) -> dict:
        """Отмена TWAP-ордера.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#cancel-a-twap-order
        """
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        action = {
            "type": "twapCancel",
            "a": asset,
            "t": twap_id,
        }

        effective_vault = vault_address or self._vault_address
        action_nonce = nonce if nonce is not None else int(time.time() * 1000)
        signature = _sign_l1_action(
            self._wallet,
            action,
            effective_vault,
            action_nonce,
            expires_after,
        )

        payload: dict[str, Any] = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }
        if effective_vault is not None:
            payload["vaultAddress"] = effective_vault
        if expires_after is not None:
            payload["expiresAfter"] = expires_after

        return await self._post_request("/exchange", data=payload)

    async def reserve_request_weight(
        self,
        weight: int,
        nonce: int | None = None,
        expires_after: int | None = None,
        vault_address: str | None = None,
    ) -> dict:
        """Резервирование дополнительного лимита действий.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#reserve-additional-actions
        """
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        action = {"type": "reserveRequestWeight", "weight": weight}

        effective_vault = vault_address or self._vault_address
        action_nonce = nonce if nonce is not None else int(time.time() * 1000)
        signature = _sign_l1_action(
            self._wallet,
            action,
            effective_vault,
            action_nonce,
            expires_after,
        )

        payload: dict[str, Any] = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }
        if effective_vault is not None:
            payload["vaultAddress"] = effective_vault
        if expires_after is not None:
            payload["expiresAfter"] = expires_after

        return await self._post_request("/exchange", data=payload)

    async def invalidate_pending_nonce(
        self,
        nonce: int | None = None,
        expires_after: int | None = None,
        vault_address: str | None = None,
    ) -> dict:
        """Инвалидация ожидающего nonce без выполнения действия.

        https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/exchange-endpoint#invalidate-pending-nonce-noop
        """
        if self._wallet is None:
            raise NotAuthorized("Private key is required for private endpoints.")

        action = {"type": "noop"}

        effective_vault = vault_address or self._vault_address
        action_nonce = nonce if nonce is not None else int(time.time() * 1000)
        signature = _sign_l1_action(
            self._wallet,
            action,
            effective_vault,
            action_nonce,
            expires_after,
        )

        payload: dict[str, Any] = {
            "action": action,
            "nonce": action_nonce,
            "signature": signature,
        }
        if effective_vault is not None:
            payload["vaultAddress"] = effective_vault
        if expires_after is not None:
            payload["expiresAfter"] = expires_after

        return await self._post_request("/exchange", data=payload)



"""Пакет, содержащий реализации клиентов и менеджеров для работы с биржей BingX."""

__all__ = [
    "Client",
    "UniClient",
    "UserWebsocket",
    "WebsocketManager",
    "UniWebsocketManager",
    "ExchangeInfo",
]

from .client import Client
from .exchange_info import ExchangeInfo
from .uni_client import UniClient
from .uni_websocket_manager import UniWebsocketManager
from .user_websocket import UserWebsocket
from .websocket_manager import WebsocketManager


async def load_exchange_info() -> None:
    """Загружает информацию о бирже BingX."""
    await ExchangeInfo.load_exchange_info()


async def start_exchange_info(parse_interval_seconds: int = 60 * 60) -> None:
    """Запускает процесс обновления информации о бирже BingX."""
    await ExchangeInfo.start(parse_interval_seconds)
